import operator

import numpy as np
import pytest
from numpy.testing import assert_equal

from pyoperators import MaskOperator, PackOperator, Rotation2dOperator
from pyoperators.core import DeletedOperator
from pyoperators.utils.testing import assert_same
from pysimulators.operators import ProjectionOperator
from pysimulators.sparse import FSRMatrix, FSRRotation2dMatrix, FSRRotation3dMatrix

from .common import BIGGEST_FLOAT_TYPE, FLOAT_TYPES, SINT_TYPES

CLASSES = FSRMatrix, FSRRotation2dMatrix, FSRRotation3dMatrix
index1 = [2, -1, 2, 1, 3, -1]
index2 = [-1, 3, 1, 1, 0, -1]
value1 = [0, 0, 0, 1.1, 0.3, 10]
value2 = [1, 2.2, 0.5, 1, 1, 10]
rotation1 = Rotation2dOperator(
    np.arange(6) * 30, degrees=True, dtype=BIGGEST_FLOAT_TYPE
)
rotation2 = Rotation2dOperator(
    np.arange(6, 12) * 30, degrees=True, dtype=BIGGEST_FLOAT_TYPE
)


def min_dtype(t1, t2):
    return min(t1, t2, key=lambda x: x().itemsize)


def get_projection(cls, itype, ftype, stype=None):
    if stype is None:
        stype = ftype

    if cls is FSRMatrix:
        dtype = [('index', itype), ('value', ftype)]
        data = np.recarray((6, 2), dtype=dtype)
        data[..., 0].index, data[..., 1].index = index1, index2
        data[..., 0].value, data[..., 1].value = value1, value2
        matrix = FSRMatrix((6, 5), data=data)
        return ProjectionOperator(matrix, dtype=stype)

    if cls is FSRRotation2dMatrix:
        dtype = [('index', itype), ('r11', ftype), ('r21', ftype)]
        data = np.recarray((6, 2), dtype=dtype)
        data[..., 0].index, data[..., 1].index = index1, index2
        data[..., 0].r11 = value1 * rotation1.data[:, 0, 0]
        data[..., 1].r11 = value2 * rotation2.data[:, 0, 0]
        data[..., 0].r21 = value1 * rotation1.data[:, 1, 0]
        data[..., 1].r21 = value2 * rotation2.data[:, 1, 0]
        matrix = FSRRotation2dMatrix((6 * 2, 5 * 2), data=data)
        return ProjectionOperator(matrix, dtype=stype)

    if cls is FSRRotation3dMatrix:
        dtype = [('index', itype), ('r11', ftype), ('r22', ftype), ('r32', ftype)]
        data = np.recarray((6, 2), dtype=dtype)
        data[..., 0].index, data[..., 1].index = index1, index2
        data[..., 0].r11, data[..., 1].r11 = value1, value2
        data[..., 0].r22 = value1 * rotation1.data[:, 0, 0]
        data[..., 1].r22 = value2 * rotation2.data[:, 0, 0]
        data[..., 0].r32 = value1 * rotation1.data[:, 1, 0]
        data[..., 1].r32 = value2 * rotation2.data[:, 1, 0]
        matrix = FSRRotation3dMatrix((6 * 3, 5 * 3), data=data)
        return ProjectionOperator(matrix, dtype=stype)

    raise


@pytest.mark.parametrize('cls', CLASSES)
@pytest.mark.parametrize('itype', SINT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
def test_kernel(cls, itype, ftype):
    expected = [False, False, True, False, True]

    proj = get_projection(cls, itype, ftype)
    actual = proj.canonical_basis_in_kernel()
    assert_equal(actual, expected)
    out = np.empty(5, bool)
    actual = proj.canonical_basis_in_kernel(out=out)
    assert_equal(actual, expected)
    kernel0 = [True, True, True, False, False]
    out = np.array(kernel0)
    proj.canonical_basis_in_kernel(out=out, operation=operator.iand)
    assert_equal(out, np.array(kernel0) & expected)
    kernel0 = [False, True, False, False, True]
    out = np.array(kernel0)
    proj.canonical_basis_in_kernel(out=out, operation=operator.ior)
    assert_equal(out, np.array(kernel0) | expected)


@pytest.mark.parametrize('cls', CLASSES)
@pytest.mark.parametrize('itype', SINT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize('vtype', FLOAT_TYPES)
def test_pT1(cls, itype, ftype, vtype):
    proj_ref = get_projection(FSRMatrix, np.int32, BIGGEST_FLOAT_TYPE)
    expected = proj_ref.T(np.ones(proj_ref.matrix.shape[0], BIGGEST_FLOAT_TYPE))
    expected_ = np.asarray(expected, min_dtype(ftype, vtype))

    proj = get_projection(cls, itype, ftype, vtype)
    pT1 = proj.pT1()
    assert_same(pT1, expected_)
    pT1 = proj.pT1(out=pT1)
    assert_same(pT1, expected_)
    pT1[...] = 2
    proj.pT1(out=pT1, operation=operator.iadd)
    assert_same(pT1, 2 + expected_)


@pytest.mark.parametrize('cls', [FSRMatrix, FSRRotation3dMatrix])
@pytest.mark.parametrize('itype', SINT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize('vtype', FLOAT_TYPES)
def test_pTx_pT1(cls, itype, ftype, vtype):
    input_fsr = np.array([1.1, 2, -1.3, 2, -3, 4.1], dtype=BIGGEST_FLOAT_TYPE)
    # input_fsrrot2d = input_fsr[:, None] * Rotation2dOperator(
    #    np.arange(6) * 30, dtype=BIGGEST_FLOAT_TYPE
    # )([1, 0])
    input_fsrrot3d = np.array(
        [input_fsr, np.random.random_sample(6), np.random.random_sample(6)]
    ).T
    inputs = {
        FSRMatrix: input_fsr,
        # FSRRotation2dMatrix: input_fsrrot2d,
        FSRRotation3dMatrix: input_fsrrot3d,
    }

    def func(cls, itype, ftype, vtype):
        proj = get_projection(cls, itype, ftype, vtype)
        input = np.asarray(inputs[cls], vtype)
        expectedx_ = np.asarray(expectedx, min_dtype(ftype, vtype))
        expected1_ = np.asarray(expected1, min_dtype(ftype, vtype))
        pTx, pT1 = proj.pTx_pT1(input)
        assert_same(pTx, expectedx_)
        assert_same(pT1, expected1_)
        proj.pTx_pT1(input, out=(pTx, pT1))
        assert_same(pTx, expectedx_)
        assert_same(pT1, expected1_)
        pTx[...] = 1
        pT1[...] = 2
        proj.pTx_pT1(input, out=(pTx, pT1), operation=operator.iadd)
        assert_same(pTx, 1 + expectedx_)
        assert_same(pT1, 2 + expected1_)

    proj_ref = get_projection(FSRMatrix, np.int32, BIGGEST_FLOAT_TYPE)
    expectedx = proj_ref.T(input_fsr)
    expected1 = proj_ref.T(np.ones(proj_ref.matrix.shape[0], BIGGEST_FLOAT_TYPE))
    func(cls, itype, ftype, vtype)


@pytest.mark.parametrize('cls', CLASSES)
@pytest.mark.parametrize('itype', SINT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize('inplace', [False, True])
def test_restrict(cls, itype, ftype, inplace):
    restriction = np.array([True, False, False, True, True])
    kernel = [False, False, True, False, True]

    proj_ref = get_projection(cls, itype, ftype)
    proj_ = get_projection(cls, itype, ftype)
    proj = proj_.restrict(restriction, inplace=inplace)
    if inplace:
        assert type(proj_) is DeletedOperator
    if cls is not FSRMatrix:

        def pack(v):
            return v[restriction, :]

    else:
        pack = PackOperator(restriction)
    masking = MaskOperator(~restriction, broadcast='rightward')
    block_size = proj_ref.matrix.block_shape[1]
    shape = (5,) + ((block_size,) if block_size > 1 else ())
    x = np.arange(5 * block_size).reshape(shape) + 1
    assert_equal(proj_ref(masking(x)), proj(pack(x)))
    pack = PackOperator(restriction)
    assert_equal(pack(kernel), proj.canonical_basis_in_kernel())



# Copyrights 2010-2011 Pierre Chanial
# All rights reserved
#

import re
import sys
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import numpy as np
from numpy.lib import NumpyVersion

from pyoperators.memory import empty

from . import _flib as flib

__all__ = ['ConversionInput', 'ConversionType', 'Quantity', 'UnitError', 'units']


numpy_version = NumpyVersion(np.__version__)
if numpy_version < '2.2.0':
    np.matvec = None
    np.vecmat = None


_RE_UNIT = re.compile(r' *([/*])? *([a-zA-Z_"\']+|\?+)(\^-?[0-9]+(\.[0-9]*)?)? *')


class UnitError(Exception):
    pass


UnitType = dict[str, float]


class ConversionInput(Enum):
    FIRST = auto()
    SECOND = auto()
    SEQUENCE = auto()
    SEQUENCE_EXCEPT_FIRST = auto()
    FIRST_SEQUENCE = auto()
    FIRST_NESTED = auto()
    SECOND_SEQUENCE = auto()
    FIRST_AND_SECOND = auto()
    FIRST_AND_THIRD = auto()
    SECOND_AND_THIRD = auto()


class ConversionOutput(Enum):
    DEFAULT = auto()  # The function simply returns an array
    SEQUENCE = auto()  # The function returns a sequence of arrays
    FIRST_IF_SEQUENCE = (
        auto()
    )  # If the output is a sequence, only convert the first element
    FIRST_IN_SEQUENCE = auto()  # Only convert the first element of the output sequence
    NONE = auto()  # Function returns None


class ConversionAction(Enum):
    SAME = auto()  # homogenize and propagate the input units to the output
    EACH = (
        auto()
    )  # when the input is *args and output is a sequence, each quantity is propagated
    SQUARE = auto()  # square the input units
    PROD = auto()  # output units is the product of the inputs
    UNITLESS = auto()  # the output is an ndarray


@dataclass
class ConversionType:
    input: ConversionInput = ConversionInput.FIRST
    output: ConversionOutput = ConversionOutput.DEFAULT
    action: ConversionAction = ConversionAction.SAME


DEFAULT_CONVERSION = ConversionType()

FUNCTIONS = {
    np.amax: DEFAULT_CONVERSION,
    np.amin: DEFAULT_CONVERSION,
    np.append: ConversionType(ConversionInput.FIRST_AND_SECOND),
    np.around: DEFAULT_CONVERSION,
    np.array_split: ConversionType(output=ConversionOutput.SEQUENCE),
    np.atleast_1d: DEFAULT_CONVERSION,
    np.atleast_2d: DEFAULT_CONVERSION,
    np.atleast_3d: DEFAULT_CONVERSION,
    np.average: DEFAULT_CONVERSION,
    # np.block: ConversionType(ConversionInput.FIRST_NESTED),
    np.broadcast_arrays: ConversionType(
        ConversionInput.SEQUENCE, ConversionOutput.SEQUENCE, ConversionAction.EACH
    ),
    np.broadcast_to: DEFAULT_CONVERSION,
    np.choose: ConversionType(ConversionInput.SECOND_SEQUENCE),
    np.clip: DEFAULT_CONVERSION,
    np.column_stack: ConversionType(ConversionInput.FIRST_SEQUENCE),
    np.compress: ConversionType(ConversionInput.SECOND),
    np.concat: ConversionType(ConversionInput.FIRST_SEQUENCE),
    np.concatenate: ConversionType(ConversionInput.FIRST_SEQUENCE),
    np.convolve: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.copy: DEFAULT_CONVERSION,
    # np.corrcoef: NO_CONVERSION,
    np.correlate: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.cov: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.cross: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.cumsum: DEFAULT_CONVERSION,
    getattr(np, 'cumulative_sum', None): DEFAULT_CONVERSION,  # Numpy 2.1
    np.delete: DEFAULT_CONVERSION,
    np.diag: DEFAULT_CONVERSION,
    np.diagflat: DEFAULT_CONVERSION,
    np.diagonal: DEFAULT_CONVERSION,
    np.diff: DEFAULT_CONVERSION,
    np.dot: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.dsplit: ConversionType(output=ConversionOutput.SEQUENCE),
    np.dstack: ConversionType(ConversionInput.FIRST_SEQUENCE),
    np.einsum: ConversionType(
        ConversionInput.SEQUENCE_EXCEPT_FIRST, action=ConversionAction.PROD
    ),
    np.expand_dims: DEFAULT_CONVERSION,
    np.extract: ConversionType(ConversionInput.SECOND),
    np.fill_diagonal: ConversionType(
        ConversionInput.FIRST_AND_SECOND, output=ConversionOutput.NONE
    ),
    np.fix: DEFAULT_CONVERSION,
    np.flip: DEFAULT_CONVERSION,
    np.fliplr: DEFAULT_CONVERSION,
    np.flipud: DEFAULT_CONVERSION,
    np.hsplit: ConversionType(output=ConversionOutput.SEQUENCE),
    np.hstack: ConversionType(ConversionInput.FIRST_SEQUENCE),
    np.imag: DEFAULT_CONVERSION,
    np.inner: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.insert: ConversionType(ConversionInput.FIRST_AND_THIRD),
    np.interp: DEFAULT_CONVERSION,
    np.intersect1d: ConversionType(ConversionInput.FIRST_AND_SECOND),
    np.kron: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.max: DEFAULT_CONVERSION,
    np.mean: DEFAULT_CONVERSION,
    np.median: DEFAULT_CONVERSION,
    np.min: DEFAULT_CONVERSION,
    np.moveaxis: DEFAULT_CONVERSION,
    np.nan_to_num: DEFAULT_CONVERSION,
    np.nancumsum: DEFAULT_CONVERSION,
    np.nanmax: DEFAULT_CONVERSION,
    np.nanmean: DEFAULT_CONVERSION,
    np.nanmedian: DEFAULT_CONVERSION,
    np.nanmin: DEFAULT_CONVERSION,
    np.nanstd: DEFAULT_CONVERSION,
    np.nansum: DEFAULT_CONVERSION,
    np.nanvar: ConversionType(action=ConversionAction.SQUARE),
    np.outer: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.pad: DEFAULT_CONVERSION,
    np.partition: DEFAULT_CONVERSION,
    np.permute_dims: DEFAULT_CONVERSION,
    np.ptp: DEFAULT_CONVERSION,
    np.ravel: DEFAULT_CONVERSION,
    np.real: DEFAULT_CONVERSION,
    np.real_if_close: DEFAULT_CONVERSION,
    np.repeat: DEFAULT_CONVERSION,
    np.reshape: DEFAULT_CONVERSION,
    np.resize: DEFAULT_CONVERSION,
    np.roll: DEFAULT_CONVERSION,
    np.rollaxis: DEFAULT_CONVERSION,
    np.rot90: DEFAULT_CONVERSION,
    np.round: DEFAULT_CONVERSION,
    np.row_stack: ConversionType(ConversionInput.FIRST_SEQUENCE),
    np.select: ConversionType(ConversionInput.SECOND_SEQUENCE),
    np.setdiff1d: ConversionType(ConversionInput.FIRST_AND_SECOND),
    np.setxor1d: ConversionType(ConversionInput.FIRST_AND_SECOND),
    np.sort: DEFAULT_CONVERSION,
    np.sort_complex: DEFAULT_CONVERSION,
    np.split: ConversionType(output=ConversionOutput.SEQUENCE),
    np.squeeze: DEFAULT_CONVERSION,
    np.stack: ConversionType(ConversionInput.FIRST_SEQUENCE),
    np.std: DEFAULT_CONVERSION,
    np.sum: DEFAULT_CONVERSION,
    np.swapaxes: DEFAULT_CONVERSION,
    np.take: DEFAULT_CONVERSION,
    np.take_along_axis: DEFAULT_CONVERSION,
    np.tensordot: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.tile: DEFAULT_CONVERSION,
    np.trace: DEFAULT_CONVERSION,
    np.transpose: DEFAULT_CONVERSION,
    np.trim_zeros: DEFAULT_CONVERSION,
    np.union1d: ConversionType(ConversionInput.FIRST_AND_SECOND),
    np.unique: ConversionType(output=ConversionOutput.FIRST_IF_SEQUENCE),
    np.unique_all: ConversionType(output=ConversionOutput.FIRST_IN_SEQUENCE),
    # np.unique_counts: NO_CONVERSION,
    # np.unique_inverse: NO_CONVERSION,
    np.unique_values: DEFAULT_CONVERSION,
    getattr(np, 'unstack', None): ConversionType(
        output=ConversionOutput.SEQUENCE
    ),  # Numpy 2.1
    np.unwrap: DEFAULT_CONVERSION,
    np.var: ConversionType(action=ConversionAction.SQUARE),
    np.vdot: ConversionType(
        ConversionInput.FIRST_AND_SECOND, action=ConversionAction.PROD
    ),
    np.vsplit: ConversionType(output=ConversionOutput.SEQUENCE),
    np.vstack: ConversionType(ConversionInput.FIRST_SEQUENCE),
    np.where: ConversionType(ConversionInput.SECOND_AND_THIRD),
}
FUNCTIONS.pop(None, None)


def _extract_unit(string: str | UnitType) -> UnitType:
    """
    Convert the input string into a unit as a dictionary.

    """
    if string is None:
        return {}

    if isinstance(string, dict):
        return string

    if not isinstance(string, str):
        raise TypeError(
            "Invalid unit type '"
            + type(string).__name__
            + "'. Expected types are string or dictionary."
        )

    string = string.strip()
    result = {}
    start = 0
    while start < len(string):
        match = _RE_UNIT.match(string, start)
        if match is None:
            raise ValueError("Unit '" + string[start:] + "' cannot be understood.")
        op = match.group(1)
        u = match.group(2)
        exp = match.group(3)
        if exp is None:
            exp = 1.0
        else:
            exp = float(exp[1:])
        if op == '/':
            exp = -exp
        result = _multiply_unit_inplace(result, u, exp)
        start = start + len(match.group(0))
    return result


def _multiply_unit_inplace(unit, key, val):
    """
    Multiply a unit as a dictionary by a non-composite one
    Unlike _divide_unit, _multiply_unit and _power_unit,
    the operation is done in place, to speed up main
    caller _extract_unit.

    """
    if key in unit:
        if unit[key] == -val:
            del unit[key]
        else:
            unit[key] += val
    else:
        unit[key] = val
    return unit


def _power_unit(unit, power):
    """
    Raise to power a unit as a dictionary

    """
    if len(unit) == 0 or power == 0:
        return {}
    result = unit.copy()
    for key in unit:
        result[key] *= power
    return result


def _multiply_unit(unit1, unit2):
    """
    Multiplication of units as dictionary.

    """
    unit = unit1.copy()
    for key, val in unit2.items():
        if key in unit:
            if unit[key] == -val:
                del unit[key]
            else:
                unit[key] += val
        else:
            unit[key] = val
    return unit


def _divide_unit(unit1, unit2):
    """
    Division of units as dictionary.

    """
    unit = unit1.copy()
    for key, val in unit2.items():
        if key in unit:
            if unit[key] == val:
                del unit[key]
            else:
                unit[key] -= val
        else:
            unit[key] = -val
    return unit


def _strunit(unit):
    """
    Convert a unit as dictionary into a string.

    """
    if len(unit) == 0:
        return ''
    result = ''
    has_pos = False

    for key, val in unit.items():
        if val >= 0:
            has_pos = True
            break

    for key, val in sorted(unit.items()):
        if val < 0 and has_pos:
            continue
        result += ' ' + key
        if val != 1:
            ival = int(val)
            if abs(val - ival) <= 1e-7 * abs(val):
                val = ival
            result += '^' + str(val)

    if not has_pos:
        return result[1:]

    for key, val in sorted(unit.items()):
        if val >= 0:
            continue
        val = -val
        result += ' / ' + key
        if val != 1:
            ival = int(val)
            if abs(val - ival) <= 1e-7 * abs(val):
                val = ival
            result += '^' + str(val)

    return result[1:]


def _grab_doc(doc, func):
    # ex: zeros.__doc__ = _grab_doc(np.zeros.__doc__, 'zeros')
    doc = doc.replace(func + '(shape, ', func + '(shape, unit=None, ')
    doc = doc.replace(
        '\n    dtype : ',
        '\n    unit : string\n        Unit of '
        'the new array, e.g. ``W``. Default is None for scalar.'
        '\n    dtype : ',
    )
    doc = doc.replace('np.' + func, 'unit.' + func)
    doc = doc.replace('\n    out : ndarray', '\n    out : Quantity')
    return doc


class Quantity(np.ndarray):
    """
    Represent a quantity, i.e. a scalar or array associated with a unit.
    if dtype is not specified, the quantity is upcasted to float64 if its dtype
    is not real nor complex

    Examples
    --------

    A Quantity can be converted to a different unit:
    >>> a = Quantity(18, 'm')
    >>> a.inunit('km')
    >>> a
    Quantity(0.018, 'km')

    A more useful conversion:
    >>> sky = Quantity(4 * np.pi, 'sr')
    >>> print(sky.tounit('deg^2'))
    41252.9612494 deg^2

    Quantities can be compared:
    >>> Quantity(0.018, 'km') > Quantity(10., 'm')
    True
    >>> np.minimum(Quantity(1, 'm'), Quantity(0.1, 'km'))
    Quantity(1.0, 'm')

    Quantities can be operated on:
    >>> time = Quantity(0.2, 's')
    >>> a / time
    Quantity(0.09, 'km / s')

    Units do not have to be standard and ? can be used as a non-standard one:
    >>> value = Quantity(1., '?/detector')
    >>> value *= Quantity(100, 'detector')
    >>> value
    Quantity(100.0, '?')

    A derived unit can be converted to a SI one:
    >>> print(Quantity(2, 'Jy').SI)
    2e-26 kg / s^2

    It is also possible to add a new derived unit:
    >>> unit['kloug'] = Quantity(3, 'kg')
    >>> print(Quantity(2, 'kloug').SI)
    6.0 kg

    In fact, a non-standard unit is not handled differently than one of
    the 7 SI units, as long as it is not in the unit table:
    >>> unit['krouf'] = Quantity(0.5, 'broug^2')
    >>> print(Quantity(1, 'krouf').SI)
    0.5 broug^2

    """

    default_dtype = np.float64
    _unit = None
    _derived_units = None

    def __new__(
        cls,
        data,
        unit=None,
        derived_units=None,
        dtype=None,
        copy=True,
        order='C',
        subok=False,
        ndmin=0,
    ):

        data = np.asanyarray(data)
        if dtype is None and data.dtype.kind == 'i':
            dtype = float

        # get a new Quantity instance (or a subclass if subok is True)
        result = np.array(
            data, dtype=dtype, copy=copy, order=order, subok=True, ndmin=ndmin
        )
        if not subok and type(result) is not cls or not isinstance(result, cls):
            result = result.view(cls)

        # set unit attribute
        if unit is not None:
            result.unit = unit

        # set derived_units attribute
        if derived_units is not None:
            result.derived_units = derived_units.copy()

        return result

    def __array_finalize__(self, obj):
        # for some numpy methods (append): the result doesn't go through
        # __new__ and obj is None. We have to set the instance attributes
        self._unit = getattr(obj, '_unit', {})
        self._derived_units = getattr(obj, '_derived_units', {})

    @property
    def __array_priority__(self):
        return 1.0 if len(self._unit) == 0 else 1.5

    def __array_function__(self, func, types, args, kwargs):
        """Handle NumPy functions on Quantity objects.

        This method intercepts NumPy function calls to provide custom behavior,
        ensuring that functions use the Quantity methods instead of going through
        ufuncs when appropriate.
        """
        conversion = FUNCTIONS.get(func)
        if conversion is None:
            return super().__array_function__(func, types, args, kwargs)

        input_units = self._get_function_input_units(args, conversion)
        output_units = self._get_function_output_units(input_units, conversion)

        converted_inputs, kwargs = self._convert_function_inputs(
            args, kwargs, output_units, conversion
        )
        converted_inputs = self._view_function_inputs(converted_inputs, conversion)
        outputs = func(*converted_inputs, **kwargs)
        if conversion.output is ConversionOutput.NONE:
            return None

        converted_outputs = self._convert_function_outputs(
            outputs, output_units, conversion
        )
        return converted_outputs

    @classmethod
    def _get_function_input_units(
        cls, arrays: Iterable[Any], conversion: ConversionType
    ) -> tuple[UnitType, ...]:
        match conversion.input:
            case ConversionInput.FIRST:
                return (cls._get_units(arrays[0]),)
            case ConversionInput.SECOND:
                return (cls._get_units(arrays[1]),)
            case ConversionInput.SEQUENCE:
                return cls._get_sequence_units(arrays)
            case ConversionInput.SEQUENCE_EXCEPT_FIRST:
                return cls._get_sequence_units(arrays[1:])
            case ConversionInput.FIRST_AND_SECOND:
                return cls._get_sequence_units([arrays[0], arrays[1]])
            case ConversionInput.SECOND_AND_THIRD:
                return cls._get_sequence_units([arrays[1], arrays[2]])
            case ConversionInput.FIRST_AND_THIRD:
                return cls._get_sequence_units([arrays[0], arrays[2]])
            case ConversionInput.FIRST_SEQUENCE:
                return cls._get_sequence_units(arrays[0])
            case ConversionInput.SECOND_SEQUENCE:
                return cls._get_sequence_units(arrays[1])
            case _:
                assert False, 'unreachable'

    @classmethod
    def _get_function_output_units(
        cls, input_units: tuple[UnitType, ...], conversion: ConversionType
    ) -> UnitType | tuple[UnitType, ...]:
        match conversion.action:
            case ConversionAction.SAME:
                return next((_ for _ in input_units if _), {})
            case ConversionAction.EACH:
                return input_units
            case ConversionAction.SQUARE:
                return _power_unit(input_units[0], 2)
            case ConversionAction.PROD:
                output_unit = input_units[0]
                for input_unit in input_units[1:]:
                    output_unit = _multiply_unit(output_unit, input_unit)
                return output_unit
            case _:
                assert False, 'unreachable'

    @classmethod
    def _convert_function_inputs(
        cls,
        arrays: Any,
        keywords: dict[str, Any],
        output_units: UnitType | Sequence[UnitType],
        conversion: ConversionType,
    ) -> tuple[Any, ...]:
        out = keywords.get('out')
        if out is not None:
            keywords['out'] = cls._view_function_input(out)

        if conversion.action is not ConversionAction.SAME or conversion.input in (
            ConversionInput.FIRST,
            ConversionInput.SECOND,
        ):
            return arrays, keywords
        assert isinstance(output_units, dict)

        common_unit = output_units
        if not common_unit:
            return arrays, keywords

        homogenized_args = []
        for iarray, array in enumerate(arrays):
            if (
                iarray == 0
                and conversion.input is ConversionInput.FIRST_SEQUENCE
                or iarray == 1
                and conversion.input is ConversionInput.SECOND_SEQUENCE
            ):
                array = tuple(
                    (
                        input.tounit(common_unit)
                        if isinstance(input, Quantity) and unit and unit != common_unit
                        else input
                    )
                    for input, unit in zip(array, cls._get_sequence_units(array))
                )
            elif (
                iarray == 0
                and conversion.input
                in (
                    ConversionInput.FIRST,
                    ConversionInput.FIRST_AND_SECOND,
                    ConversionInput.FIRST_AND_THIRD,
                )
                or iarray == 1
                and conversion.input
                in (
                    ConversionInput.SECOND,
                    ConversionInput.FIRST_AND_SECOND,
                    ConversionInput.SECOND_AND_THIRD,
                )
                or iarray == 2
                and conversion.input
                in (ConversionInput.FIRST_AND_THIRD, ConversionInput.SECOND_AND_THIRD)
                or iarray > 0
                and conversion.input is ConversionInput.SEQUENCE_EXCEPT_FIRST
                or conversion.input is ConversionInput.SEQUENCE
            ):
                if (
                    isinstance(array, Quantity)
                    and array._unit
                    and array._unit != common_unit
                ):
                    array = array.tounit(common_unit)
            homogenized_args.append(array)
        return tuple(homogenized_args), keywords

    @classmethod
    def _view_function_inputs(
        cls, arrays: Any, conversion: ConversionType
    ) -> tuple[Any, ...]:
        viewed_arrays = []
        for iarray, array in enumerate(arrays):
            if (
                iarray == 0
                and conversion.input is ConversionInput.FIRST_SEQUENCE
                or iarray == 1
                and conversion.input is ConversionInput.SECOND_SEQUENCE
            ):
                array = cls._view_function_input_iterable(array)
            else:
                array = cls._view_function_input(array)
            viewed_arrays.append(array)
        return tuple(viewed_arrays)

    @classmethod
    def _view_function_input(cls, array):
        if isinstance(array, np.ndarray) and type(array) is not np.ndarray:
            return array.view(np.ndarray)
        return array

    @classmethod
    def _view_function_input_iterable(cls, arrays: Iterable[Any]) -> list[Any]:
        return [cls._view_function_input(array) for array in arrays]

    def _convert_function_outputs(
        self,
        arrays: Any,
        output_units: UnitType | Sequence[UnitType],
        conversion: ConversionType,
    ) -> Any | tuple[Any, ...]:
        if np.isscalar(arrays) or isinstance(arrays, np.ndarray):
            # covers the case DEFAULT and FIRST_IF_SEQUENCE when then the output is not
            # a sequence
            if np.isscalar(arrays):
                output = type(self)(arrays)
            else:
                output = arrays.view(type(self))
            output.__array_finalize__(self)
            output._unit = output_units
            return output

        if isinstance(output_units, dict):
            output_units = len(arrays) * [output_units]

        # Check if arrays is a namedtuple
        is_namedtuple = (
            isinstance(arrays, tuple)
            and hasattr(arrays, '_fields')
            and hasattr(arrays, '_make')
        )

        outputs = []
        for iarray, (array, units) in enumerate(zip(arrays, output_units)):
            if (
                iarray == 0
                and conversion.output
                in (
                    ConversionOutput.FIRST_IF_SEQUENCE,
                    ConversionOutput.FIRST_IN_SEQUENCE,
                )
                or conversion.output is ConversionOutput.SEQUENCE
            ):
                array = array.view(type(self))
                array.__array_finalize__(self)
                array._unit = units
            outputs.append(array)

        # If input was a namedtuple, reconstruct it with the same type
        if is_namedtuple:
            return type(arrays)(*outputs)

        return outputs

    @classmethod
    def _get_sequence_units(cls, arrays) -> tuple[UnitType, ...]:
        return tuple(cls._get_units(_) for _ in arrays)

    @classmethod
    def _get_units(cls, array) -> UnitType:
        if not isinstance(array, Quantity):
            return {}
        return array._unit

    @classmethod
    def _get_inputs_with_common_units(
        cls,
        arrays: Sequence[Any, ...],
    ) -> tuple[Any, UnitType]:
        input_units = cls._get_sequence_units(arrays)
        common_unit = next((_ for _ in input_units if _), {})
        inputs = tuple(
            input.tounit(common_unit) if unit and unit != common_unit else input
            for input, unit in zip(arrays, input_units)
        )
        return inputs, common_unit

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Handle ufunc operations on Quantity objects."""
        # Get the output if provided
        out = kwargs.get('out', None)

        input_units = self._get_sequence_units(inputs)

        if ufunc in (
            np.add,
            np.clip,
            np.fmod,
            np.hypot,
            np.maximum,
            np.minimum,
            np.subtract,
        ):
            inputs, output_unit = self._get_inputs_with_common_units(inputs)

        elif ufunc in (
            np.equal,
            np.greater,
            np.greater_equal,
            np.less,
            np.less_equal,
            np.not_equal,
        ):
            inputs, _ = self._get_inputs_with_common_units(inputs)
            output_unit = None

        elif ufunc in (
            np.absolute,
            np.ceil,
            np.conjugate,
            np.copysign,
            np.fabs,
            np.float_power,
            np.floor,
            np.fmax,
            np.fmin,
            np.hypot,
            np.negative,
            np.rint,
            np.trunc,
        ):
            output_unit = self._unit

        elif ufunc is np.reciprocal:
            output_unit = _power_unit(input_units[0], -1)

        elif ufunc is np.cbrt:
            output_unit = _power_unit(input_units[0], 1 / 3)

        elif ufunc is np.sqrt:
            output_unit = _power_unit(input_units[0], 0.5)

        elif ufunc is np.square:
            output_unit = _power_unit(input_units[0], 2)

        elif ufunc is np.power:
            power = (
                inputs[1].view(np.ndarray)
                if isinstance(inputs[1], Quantity)
                else inputs[1]
            )
            output_unit = _power_unit(input_units[0], power)

        elif ufunc in (np.matmul, np.matvec, np.multiply, np.vecdot, np.vecmat):
            output_unit = _multiply_unit(input_units[0], input_units[1])

        elif ufunc in (np.floor_divide, np.true_divide, np.divide):
            output_unit = _divide_unit(input_units[0], input_units[1])

        elif ufunc in (np.exp, np.expm1, np.exp2, np.log, np.log10, np.log2, np.log1p):
            # Transcendental functions return dimensionless results
            output_unit = {}

        else:
            output_unit = None

        # Process outputs if provided
        # TODO: check me
        if out is not None:
            kwargs['out'] = tuple(
                o.view(np.ndarray) if isinstance(o, Quantity) else o
                for o in (out if isinstance(out, tuple) else (out,))
            )

        # Call the ufunc
        result = super().__array_ufunc__(
            ufunc,
            method,
            *[
                input.view(np.ndarray) if isinstance(input, Quantity) else input
                for input in inputs
            ],
            **kwargs,
        )
        if result is None or result is NotImplemented or output_unit is None:
            return result

        # Handle tuple results (for ufuncs with multiple outputs)
        if isinstance(result, tuple):
            return tuple(self._process_ufunc_result(r, output_unit) for r in result)

        return self._process_ufunc_result(result, output_unit)

    def _process_ufunc_result(self, result, output_unit):
        """
        Process the result of a ufunc call, setting the appropriate unit.
        """
        if np.isscalar(result):
            result = type(self)(result)
        else:
            result = result.view(type(self))
        result.__array_finalize__(self)
        result._unit = output_unit
        return result

    #    def __getattr__(self, name):
    #        if self.dtype.names is None or name not in self.dtype.names:
    #            raise AttributeError(
    #                "'" + self.__class__.__name__ + "' object has"
    #                " no attribute '" + name + "'"
    #            )
    #        return self[name]

    def __getitem__(self, key):
        """
        x.__getitem__(y) <==> x[y]

        Return the item described by the key.

        """
        if key is Ellipsis:
            return self

        item = np.ndarray.__getitem__(self, key)
        if not isinstance(item, np.ndarray):
            return item
        if self.dtype.kind == 'V' and isinstance(key, str):
            return item.view(Quantity)

        if isinstance(key, list):
            key = tuple(key)

        if not isinstance(key, tuple):
            key = (key,)

        # Replace Ellipsis with ':'
        try:
            pos = key.index(Ellipsis)
            key = key[:pos] + (self.ndim - len(key)) * (slice(None),) + key[pos + 1 :]
        except ValueError:
            pass

        key += (self.ndim - len(key)) * (slice(None),)

        # update the broadcastable derived units
        du = None
        for d, v in item._derived_units.items():
            try:
                pos = d.index('[')
            except ValueError:
                continue

            # the derived unit is broadcastable, let's update it
            broadcast = d[pos + 1 : -1]
            if broadcast == 'leftward':
                v = v[key[-v.ndim :]]
            else:
                v = v[key[: v.ndim]]

            if du is None:
                du = item._derived_units.copy()

            # if the derived unit has become a scalar, remove the brackets
            if v.ndim == 0:
                del du[d]
                du[d[:pos]] = v
            else:
                du[d] = v

        if du is not None:
            item._derived_units = du

        return item

    def __getslice__(self, i, j):
        """
        x.__getslice__(i, j) <==> x[i:j]

        Return the slice described by (i, j).  The use of negative
        indices is not supported.

        """
        return self.__getitem__(slice(i, j))

    @property
    def magnitude(self):
        """
        x.magnitude <==> x.view(np.ndarray)

        Return the magnitude of the quantity
        """
        return self.view(np.ndarray)

    @magnitude.setter
    def magnitude(self, value):
        """
        Set the magnitude of the quantity
        """
        if self.ndim == 0:
            self.shape = (1,)
            try:
                self.view(np.ndarray)[:] = value
            except ValueError as error:
                raise error
            finally:
                self.shape = ()
        else:
            self.view(np.ndarray)[:] = value

    @property
    def unit(self):
        """
        Return the Quantity unit as a string

        Example
        -------
        >>> Quantity(32., 'm/s').unit
        'm / s'

        """
        return _strunit(self._unit)

    @unit.setter
    def unit(self, unit):
        self._unit = _extract_unit(unit)

    @property
    def derived_units(self):
        """
        Return the derived units associated with the quantity

        """
        return self._derived_units

    @derived_units.setter
    def derived_units(self, derived_units):
        if derived_units is not None:
            if not isinstance(derived_units, dict):
                raise TypeError('Input derived units are not a dict.')
            for key in derived_units:
                if not isinstance(derived_units[key], Quantity) and not hasattr(
                    derived_units[key], '__call__'
                ):
                    raise UnitError(
                        "The user derived unit '" + key + "' is not a Quantity."
                    )
                try:
                    pos = key.index('[')
                except ValueError:
                    continue
                if key[pos:] not in ('[leftward]', '[rightward]'):
                    raise UnitError(
                        f'Invalid broadcast : {key[pos:]!r}. Valid values a'
                        f"re '[leftward]' or '[rightward]'."
                    )
        self._derived_units = derived_units

    def tounit(self, unit: str | UnitType) -> Self:
        """
        Convert a Quantity into a new unit

        Parameters
        ----------
        unit : string
             A string representing the unit into which the Quantity
             should be converted

        Returns
        -------
        res : Quantity
            A shallow copy of the Quantity in the new unit

        Example
        -------
        >>> q = Quantity(1., 'km')
        >>> p = q.tounit('m')
        >>> print(p)
        1000.0 m
        """
        result = self.copy()
        result.inunit(unit)
        return result

    def inunit(self, unit: str | UnitType) -> None:
        """
        In-place conversion of a Quantity into a new unit

        Parameters
        ----------
        unit : string
             A string representing the unit into which the Quantity
             should be converted

        Example
        -------
        >>> q = Quantity(1., 'km')
        >>> q.inunit('m')
        >>> print(q)
        1000.0 m
        """

        newunit = _extract_unit(unit)

        if len(self._unit) == 0 or len(newunit) == 0:
            self._unit = newunit
            return

        # the header may contain information used to do unit conversions
        if hasattr(self, '_header'):
            q1 = self.__class__(
                1,
                header=self._header,
                unit=self._unit,
                derived_units=self.derived_units,
            ).SI
            q2 = self.__class__(
                1, header=self._header, unit=newunit, derived_units=self.derived_units
            ).SI
        else:
            q1 = Quantity(1.0, self._unit, self.derived_units).SI
            q2 = Quantity(1.0, newunit, self.derived_units).SI

        if q1._unit != q2._unit:
            raise UnitError(
                "Units '"
                + self.unit
                + "' and '"
                + _strunit(newunit)
                + "' are incompatible."
            )
        factor = q1.magnitude / q2.magnitude
        if self.ndim == 0:
            self.magnitude *= factor
        else:
            self.magnitude.T[:] *= factor.T
        self._unit = newunit

    @property
    def SI(self):
        """
        Return the quantity in SI unit. If the quantity has
        no units, the quantity itself is returned, otherwise, a
        shallow copy is returned.

        Example
        -------
        >>> print(Quantity(1., 'km').SI)
        1000.0 m
        """
        if len(self._unit) == 0:
            return self

        ffast = Quantity(1.0, '')
        fslow = Quantity(1.0, '')
        for key, val in self._unit.items():

            # check if the unit is a local derived unit
            newfactor, broadcast = _check_du(self, key, val, self.derived_units)

            # check if the unit is a global derived unit
            if newfactor is None:
                newfactor, broadcast = _check_du(self, key, val, units)

            # if the unit is not derived, we add it to the dictionary
            if newfactor is None:
                _multiply_unit_inplace(fslow._unit, key, val)
                continue

            # factor may be broadcast
            if broadcast == 'leftward':
                fslow = fslow * newfactor
            else:
                ffast = (ffast.T * newfactor.T).T

        result = self * fslow.magnitude
        result = (result.T * ffast.magnitude.T).T
        result._unit = _multiply_unit(fslow._unit, ffast._unit)

        return result

    def __reduce__(self):
        state = list(np.ndarray.__reduce__(self))
        try:
            subclass_state = self.__dict__.copy()
        except AttributeError:
            subclass_state = {}
        for cls in self.__class__.__mro__:
            for slot in cls.__dict__.get('__slots__', ()):
                try:
                    subclass_state[slot] = getattr(self, slot)
                except AttributeError:
                    pass
        state[2] = (state[2], subclass_state)
        return tuple(state)

    def __setstate__(self, state):
        ndarray_state, subclass_state = state
        np.ndarray.__setstate__(self, ndarray_state)
        for k, v in subclass_state.items():
            setattr(self, k, v)

    def __repr__(self):
        return (
            type(self).__name__
            + '('
            + str(np.asarray(self))
            + ", '"
            + _strunit(self._unit)
            + "')"
        )

    def __str__(self):
        result = str(np.asarray(self))
        if len(self._unit) == 0:
            return result
        return result + ' ' + _strunit(self._unit)

    @classmethod
    def empty(
        cls, shape, unit=None, derived_units=None, dtype=None, order=None, **keywords
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            empty(shape, dtype, order),
            dtype=dtype,
            unit=unit,
            derived_units=derived_units,
            copy=False,
            **keywords,
        )

    @classmethod
    def ones(
        cls, shape, unit=None, derived_units=None, dtype=None, order=None, **keywords
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            np.ones(shape, dtype, order),
            dtype=dtype,
            unit=unit,
            derived_units=derived_units,
            copy=False,
            **keywords,
        )

    @classmethod
    def zeros(
        cls, shape, unit=None, derived_units=None, dtype=None, order=None, **keywords
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            np.zeros(shape, dtype, order),
            dtype=dtype,
            unit=unit,
            derived_units=derived_units,
            copy=False,
            **keywords,
        )


def _check_du(input, key, val, derived_units):
    if len(derived_units) == 0:
        return None, None
    if (key, val) in derived_units:
        du, broadcast = _get_du(input, (key, val), derived_units)
        return (None, None) if du is None else (du.SI, broadcast)
    if (key, -val) in derived_units:
        du, broadcast = _get_du(input, (key, -val), derived_units)
        return (None, None) if du is None else ((1 / du).SI, broadcast)
    du, broadcast = _get_du(input, key, derived_units)
    if du is None:
        return None, None
    if val == 1.0:
        return du.SI, broadcast
    if val == -1.0:
        return (1 / du).SI, broadcast
    return (du**val).SI, broadcast


def _get_du(input, key, derived_units):
    try:
        du, broadcast = _check_in_du(key, derived_units)
    except ValueError:
        return None, None
    if isinstance(du, Callable):
        du = du(input)
    if du is None:
        return None, None
    if hasattr(input, '_header'):
        du = du.view(type(input))
        du._header = input._header
    du._derived_units = input._derived_units
    return du, broadcast


def _check_in_du(key, derived_units):
    for du, value in derived_units.items():
        try:
            pos = du.index('[')
            if key == du[:pos]:
                return value, du[pos + 1 : -1]
        except ValueError:
            if key == du:
                return value, 'leftward'
    raise ValueError


def pixel_to_pixel_reference(input):
    """
    Returns the pixel area in units of reference pixel.
    """
    if not hasattr(input, 'header'):
        return Quantity(1, 'pixel_reference')
    required = 'CRPIX,CRVAL,CTYPE'.split(',')
    keywords = np.concatenate(
        [
            (lambda i: [r + str(i + 1) for r in required])(i)
            for i in range(input.header['NAXIS'])
        ]
    )
    if not all([k in input.header for k in keywords]):
        return Quantity(1, 'pixel_reference')

    scale, status = flib.wcsutils.projection_scale(
        str(input.header).replace('\n', ''),
        input.header['NAXIS1'],
        input.header['NAXIS2'],
    )
    if status != 0:
        raise RuntimeError()

    return Quantity(scale.T, 'pixel_reference', copy=False)


def pixel_reference_to_solid_angle(input):
    """
    Returns the reference pixel area in the units defined by the FITSheader
    """
    if not hasattr(input, 'header'):
        return None

    header = input.header
    if all([key in header for key in ('CD1_1', 'CD2_1', 'CD1_2', 'CD2_2')]):
        cd = np.array(
            [[header['cd1_1'], header['cd1_2']], [header['cd2_1'], header['cd2_2']]]
        )
        area = abs(np.linalg.det(cd))
    elif 'CDELT1' in header and 'CDELT2' in header:
        area = abs(header['CDELT1'] * header['CDELT2'])
    else:
        return None

    cunit1 = header['CUNIT1'] if 'CUNIT1' in header else 'deg'
    cunit2 = header['CUNIT2'] if 'CUNIT2' in header else 'deg'
    return area * Quantity(1, cunit1) * Quantity(1, cunit2)


units_table = {
    # SI
    'A': None,
    'cd': None,
    'K': None,
    'kg': None,
    'm': None,
    'mol': None,
    'rad': None,
    's': None,
    'sr': None,
    # angle
    "'": Quantity(1, 'arcmin'),
    '"': Quantity(1, 'arcsec'),
    'arcmin': Quantity(1 / 60.0, 'deg'),
    'arcsec': Quantity(1 / 3600.0, 'deg'),
    'deg': Quantity(np.pi / 180.0, 'rad'),
    # flux_densities
    'uJy': Quantity(1.0e-6, 'Jy'),
    'mJy': Quantity(1.0e-3, 'Jy'),
    'Jy': Quantity(1.0e-26, 'W/Hz/m^2'),
    'MJy': Quantity(1.0e6, 'Jy'),
    # force
    'N': Quantity(1.0, 'kg m / s^2'),
    # frequency
    'Hz': Quantity(1.0, 's^-1'),
    # energy
    'J': Quantity(1.0, 'kg m^2 / s^2'),
    # length
    'AU': Quantity(149597870700.0, 'm'),
    'km': Quantity(1000.0, 'm'),
    'mm': Quantity(1e-3, 'm'),
    'um': Quantity(1e-6, 'm'),
    'pc': Quantity(30.857e15, 'm'),
    # power
    'W': Quantity(1.0, 'kg m^2 / s^3'),
    # pressure
    'atm': Quantity(101325.0, 'Pa'),
    'bar': Quantity(1e5, 'Pa'),
    'mmHg': Quantity(1 / 760, 'atm'),
    'cmHg': Quantity(10, 'mmHg'),
    'Pa': Quantity(1, 'kg / m / s^2'),
    'Torr': Quantity(1, 'mmHg'),
    # resistivity
    'Ohm': Quantity(1.0, 'V/A'),
    # solid angle
    ('rad', 2): Quantity(1.0, 'sr'),
    'pixel': pixel_to_pixel_reference,
    'pixel_reference': pixel_reference_to_solid_angle,
    # time
    'ms': Quantity(1e-3, 's'),
    'us': Quantity(1e-6, 's'),
    # misc
    'C': Quantity(1.0, 'A s'),
    'V': Quantity(1.0, 'kg m^2 / A / s^3'),
}


class Unit(dict):
    def __init__(self):
        for k, v in units_table.items():
            self[k] = v
            if isinstance(k, str):
                setattr(self, k, Quantity(1, k))


units = Unit()



import numpy as np
from numpy.testing import assert_equal

from pysimulators import FitsArray, Map, Quantity, Tod

SINT_TYPES = [np.int8, np.int16, np.int32, np.int64]
UINT_TYPES = [np.uint8, np.uint16, np.uint32, np.uint64]
INT_TYPES = SINT_TYPES + UINT_TYPES
FLOAT_TYPES = [np.float16, np.float32, np.float64]
if hasattr(np, 'float128'):
    FLOAT_TYPES.append(np.float128)
    BIGGEST_FLOAT_TYPE = np.float128
else:
    BIGGEST_FLOAT_TYPE = np.float64


def assert_equal_subclass(x, y):
    assert type(x) is type(y)
    assert x.dtype == y.dtype
    assert_equal(x.view(np.ndarray), y.view(np.ndarray))

    if isinstance(x, Quantity):
        assert x._unit == y._unit
        assert x._derived_units == y._derived_units

    if isinstance(x, FitsArray):
        assert x.header == y.header

    if isinstance(x, Map):
        if x.coverage is None:
            assert y.coverage is None
        else:
            assert y.coverage is not None
            assert x.coverage.shape == y.coverage.shape
            assert_equal(x.coverage, y.coverage)
        if x.error is None:
            assert y.error is None
        else:
            assert y.error is not None
            assert x.error.shape == y.error.shape
            assert_equal(x.error, y.error)

    if isinstance(x, Tod):
        if x.mask is None:
            assert y is None
        else:
            assert y.mask is not None
            assert x.mask.shape == y.mask.shape
            assert_equal(x.mask, y.mask)



"""
Define PackedTable classes:

PackedTable:
  - Layout
      - LayoutGrid
          - LayoutGridSquare
  - Sampling
      - Pointing
          - PointingEquatorial
          - PointingHorizontal
  - Scene
      - SceneGrid
      - SceneHealpix

"""

import copy
import functools
import inspect
import types
import warnings
from collections.abc import Callable

import numpy as np

from pyoperators import MPI
from pyoperators.utils import (
    ilast_is_not,
    isalias,
    isclassattr,
    isscalarlike,
    product,
    split,
    tointtuple,
)

from ..warnings import PySimulatorsWarning

__all__ = ['PackedTable']


class PackedTable:
    """
    The PackedTable class gathers information from a set of elements which can
    have a multi-dimensional layout. This information can transparently be
    accessed as packed or unpacked arrays.

    Example
    -------
    Let's consider a 3x3 array of detectors, in which the top-left detector
    is not working. We will define a mask to flag this detector.
    >>> selection = [[True, True, False],
    ...              [True, True, True],
    ...              [True, True, True]]
    >>> gain = [[1.0, 1.2, 1.5],
    ...         [0.9, 1.0, 1.0],
    ...         [0.8, 1.0, 1.0]]
    >>> table = PackedTable((3, 3), selection=selection, gain=gain)

    Only the values for the selected detectors are stored, in 1-dimensional
    arrays:
    >>> table.gain
    array([ 1. ,  1.2,  0.9,  1. ,  1. ,  0.8,  1. ,  1. ])

    But the 2-dimensional table can be recovered:
    >>> table.all.gain
    array([[ 1. ,  1.2,  nan],
           [ 0.9,  1. ,  1. ],
           [ 0.8,  1. ,  1. ]])

    The number of selected detectors is:
    >>> len(table)
    8

    and the number of all detectors is:
    >>> len(table.all)
    9

    Now, let's have a more complex example: an array of detectors made of
    4 identical 3x3 subarrays in which one corner detector is blind and
    for which we will define an indexing scheme.
    The first subarray is placed on the upper right quadrant and the position
    of the other arrays is obtained by rotating the first array by 90, 180 and
    270 degrees.
    A natural indexing would be given by:

    >>> ordering = [[-1, 14, 17,  0,  1, -1],
    ...             [10, 13, 16,  3,  4,  5],
    ...             [ 9, 12, 15,  6,  7,  8],
    ...             [26, 25, 24, 33, 30, 27],
    ...             [23, 22, 21, 34, 31, 28],
    ...             [-1, 19, 18, 35, 32, -1]]

    The following mask only keeps the 2 subarrays on the left:

    >>> selection = [[ True,  True,  True, False, False, False],
    ...              [ True,  True,  True, False, False, False],
    ...              [ True,  True,  True, False, False, False],
    ...              [ True,  True,  True, False, False, False],
    ...              [ True,  True,  True, False, False, False],
    ...              [ True,  True,  True, False, False, False]]
    >>> table = PackedTable((6, 6), selection=selection, ordering=ordering)

    Then, the numbering of the table fields follows the list of selected
    indices stored in:

    >>> print(table.index)
    [12  6 13  7  1 14  8  2 32 31 26 25 24 20 19 18]

    which are the 1d-collapsed indices of the following array coordinates:

    >>> print([(i // 6, i % 6) for i in table.index])
    [(2, 0), (1, 0), (2, 1), (1, 1), (0, 1), (2, 2), (1, 2), (0, 2),
     (5, 2), (5, 1), (4, 2), (4, 1), (4, 0), (3, 2), (3, 1), (3, 0)]

    """

    def __init__(self, shape, selection=None, ordering=None, ndim=None, **keywords):
        """
        shape : tuple of int
            The shape of the unpacked table attributes. For 2-dimensional
            attributes, the shape would be (nrows, ncolumns).
        ndim : int, optional
            The number of splittable (indexable) dimensions. It is the actual
            number of dimensions of the layout. It can be lower than that
            specified by the layout shape, in which case the extra dimensions
            are instructed not to be split.
        selection : array-like of bool or int, slices, optional
            The slices or the integer or boolean selection that specifies
            the selected components (and reject those that are not physically
            present or those not handled by the current MPI process when the
            table is distributed in a parallel processing).
        ordering : array-like of int, optional
            The values in this array specify an ordering of the components. It is
            used to define the 1-dimensional indexing of the packed components.
            A negative value means that the component is removed.

        """
        shape = tointtuple(shape)
        if ndim is None:
            ndim = len(shape)
        elif ndim < 1 or ndim > len(shape):
            raise ValueError(f"Invalid ndim value '{ndim}'.")
        object.__setattr__(self, 'shape', shape)
        object.__setattr__(self, 'ndim', ndim)
        object.__setattr__(self, 'comm', MPI.COMM_SELF)
        self._dtype_index = np.int64
        self._index = self._get_index(selection, ordering)

        for k, v in keywords.items():
            self._special_attributes += (k,)
            if v is None and isclassattr(k, type(self)):
                continue
            if not isinstance(v, Callable):
                v = self.pack(v, copy=True)
            setattr(self, k, v)

    _reserved_attributes = ('all', 'comm', 'index', 'ndim', 'removed', 'shape')
    _special_attributes = ('index', 'removed')

    @property
    def all(self):
        """
        Give access to the unpacked attributes, without creating a cyclic
        reference.
        """
        return UnpackedTable(self)

    @property
    def index(self):
        """Return the index as an int array."""
        index = self._index
        if index is Ellipsis:
            return np.arange(self._size_actual, dtype=self._dtype_index)
        if isinstance(index, slice):
            return np.arange(
                index.start, index.stop, index.step, dtype=self._dtype_index
            )
        return index

    @property
    def _indexable(self):
        # slice index with negative stop values cannot be used for indexing
        index = self._index
        if isinstance(index, slice) and index.stop < 0:
            return slice(index.start, None, index.step)
        return index

    @property
    def removed(self):
        return np.array(False)

    @property
    def _shape_actual(self):
        return self.shape[: self.ndim]

    @property
    def _size_actual(self):
        return product(self._shape_actual)

    def _get_index(self, selection, ordering):
        selection = self._normalize_selection(selection, self._shape_actual)

        if ordering is not None:
            ordering = np.asarray(ordering)
            if ordering.shape != self._shape_actual:
                raise ValueError('Invalid ordering dimensions.')

        if ordering is None:
            # assuming row major storage for the ordering
            if selection is Ellipsis or isinstance(selection, slice):
                return selection
            if isinstance(selection, np.ndarray) and selection.dtype == int:
                out = np.asarray(selection, self._dtype_index)
                return self._normalize_int_selection(out, self._size_actual)
            if isinstance(selection, tuple):
                s = selection[0]
                if isinstance(s, slice) and (
                    self.ndim == 1 or len(selection) == 1 and s.step == 1
                ):
                    n = product(self._shape_actual[1:])
                    return slice(s.start * n, s.stop * n, s.step)
                selection = self._selection2bool(selection)

            return self._normalize_int_selection(
                np.asarray(np.where(selection.ravel())[0], dtype=self._dtype_index),
                self._size_actual,
            )

        if selection is not Ellipsis and (
            not isinstance(selection, np.ndarray) or selection.dtype == int
        ):
            selection = self._selection2bool(selection)

        if selection is not Ellipsis:
            ordering = ordering.copy()
            ordering[~selection] = -1

        npacked = np.sum(ordering >= 0)
        if npacked == 0:
            return np.array([], self._dtype_index)
        out = np.asarray(
            np.argsort(ordering.ravel(), kind='mergesort')[-npacked:], self._dtype_index
        )
        return self._normalize_int_selection(out, self._size_actual)

    def __len__(self):
        """Return the number of actual elements in the table."""
        index = self._index
        if index is Ellipsis:
            return self._size_actual
        if isinstance(index, slice):
            return (index.stop - index.start) // index.step
        return index.size

    def __setattr__(self, key, value):
        if key in self._reserved_attributes:
            raise AttributeError(f'The attribute {key!r} is not writeable.')
        if (
            value is not None
            and not isinstance(value, Callable)
            and key in self._special_attributes
        ):
            value = np.asanyarray(value)
            if value.ndim > 0 and value.shape[0] != len(self):
                raise ValueError(
                    f"The shape '{value.shape}' is invalid. The expected first "
                    f"dimension is '{len(self)}'."
                )
            elif value.ndim == 0:
                try:
                    old_value = object.__getattribute__(self, key)
                    if old_value is not None and not isinstance(old_value, Callable):
                        old_value[...] = value
                        return
                except AttributeError:
                    pass
        object.__setattr__(self, key, value)

    def __getattribute__(self, key):
        if key == 'packed':
            warnings.warn(
                "Please update your code: the 'packed' attribute is not requi"
                'red anymore: the attributes are already packed. To access th'
                "e unpacked ones, use the 'all' attribute.",
                PySimulatorsWarning,
            )
            return self
        value = object.__getattribute__(self, key)
        if key not in object.__getattribute__(
            self, '_special_attributes'
        ) or not isinstance(value, Callable):
            return value
        spec = inspect.getfullargspec(value)
        nargs = len(spec.args)
        if spec.defaults is not None:
            nargs -= len(spec.defaults)
        if isinstance(value, types.MethodType):
            if nargs == 1:
                return value()
            return value
        if nargs == 0:
            return value()
        if nargs == 1:
            return value(self)
        return functools.partial(value, self)

    def __getitem__(self, selection):
        selection = self._normalize_selection(selection, (len(self),))
        if selection is Ellipsis:
            return self

        index = self._index
        if index is Ellipsis and (
            isinstance(selection, slice)
            or isinstance(selection, np.ndarray)
            and selection.dtype == int
        ):
            index = selection
        elif isinstance(index, slice) and isinstance(selection, slice):
            n = (selection.stop - selection.start) // selection.step
            start = index.start + selection.start * index.step
            step = index.step * selection.step
            stop = start + n * step
            index = slice(start, stop, step)
        else:
            index = self.index
            index = index[selection]

        if isinstance(index, np.ndarray):
            index = self._normalize_int_selection(index, self._size_actual)

        out = copy.copy(self)
        out._index = index
        for k in out._special_attributes:
            if k in out._reserved_attributes:
                continue
            try:
                v = self.__dict__[k]
            except KeyError:
                continue
            if isinstance(v, np.ndarray) and v.ndim > 0:
                setattr(out, k, v[selection])
        return out

    def __str__(self):
        out = type(self).__name__ + f'({self._shape_actual}, '
        attributes = ['index'] + sorted(
            _
            for _ in self._special_attributes
            if _ not in ('index', 'removed')
            and not isinstance(
                object.__getattribute__(self, _),
                (types.FunctionType, types.MethodType, type(None)),
            )
        )
        if len(attributes) > 1:
            out += '\n    '
        out += (
            ',\n    '.join(f'{_}={str(getattr(self, _))[:65]}' for _ in attributes)
            + ')'
        )
        return out

    __repr__ = __str__

    def copy(self):
        """
        Return a deep copy of the packed array.

        """
        out = copy.deepcopy(self)
        out.comm = self.comm.Dup()
        return out

    def pack(self, x, out=None, copy=False):
        """
        Convert a multi-dimensional array into a 1-dimensional array which
        only includes the selected components, potentially ordered according
        to a given ordering.

        Parameters
        ----------
        x : ndarray
            Array to be packed, whose first dimensions are equal to those
            of the table attributes.
        out : ndarray, optional
            Placeholder for the packed array.
        copy : boolean, optional
            Setting this keyword to True ensures that the output is not a view
            of the input x.

        Returns
        -------
        output : ndarray
            Packed array, whose first dimension is the number of non-removed
            components.

        See Also
        --------
        unpack : inverse method.

        Notes
        -----
        Unless the 'copy' keyword is set to True, this method does not make
        a copy if it simply does a reshape (when all components are selected
        and no ordering is specified).

        """
        if x is None:
            return None
        x = np.asanyarray(x)
        if x.ndim == 0:
            return x
        if x.shape[: self.ndim] != self._shape_actual:
            raise ValueError(
                f"Invalid unpacked shape '{x.shape}'. The first dimension(s) must be "
                f"'{self._shape_actual}'."
            )
        index = self._indexable

        packed_shape = (len(self),) + x.shape[self.ndim :]
        flat_shape = (self._size_actual,) + x.shape[self.ndim :]
        if out is not None:
            if not isinstance(out, np.ndarray):
                raise TypeError('The output array is not an ndarray.')
            if out.shape != packed_shape:
                raise ValueError(
                    f"The output array shape '{out.shape}' is invalid. The expected "
                    f"shape is '{packed_shape}'."
                )
        else:
            if not copy:
                if index is Ellipsis:
                    return x.reshape(packed_shape)
                elif isinstance(index, slice):
                    return x.reshape(flat_shape)[index]
            out = np.empty(packed_shape, dtype=x.dtype).view(type(x))
            if out.__array_finalize__ is not None:
                out.__array_finalize__(x)
        if index is Ellipsis:
            out[...] = x.reshape(packed_shape)
            return out
        x_ = x.view(np.ndarray).reshape(flat_shape)
        if isinstance(index, slice):
            out[...] = x_[index]
        else:
            np.take(x_, index, axis=0, out=out)
        return out

    def scatter(self, comm=None):
        """
        MPI-scatter of the table.

        Parameter
        ---------
        comm : MPI.Comm
            The MPI communicator of the group of processes in which the table
            will be scattered.

        """
        if self.comm.size > 1:
            raise ValueError('The table is already distributed.')
        if comm is None:
            comm = MPI.COMM_WORLD
        if comm.size == 1:
            return self

        selection = split(len(self), comm.size, comm.rank)
        out = self[selection]
        for k in out._special_attributes:
            if k in out._reserved_attributes:
                continue
            try:
                v = out.__dict__[k]
            except KeyError:
                continue
            if isinstance(v, np.ndarray) and v.ndim > 0:
                setattr(out, k, v.copy())
        object.__setattr__(out, 'comm', comm)
        return out

    def gather(self, *args):
        """
        MPI-gather the (already scattered) table or a given array.

        table_global = table_local.gather()
        array_global = table_local.gather(array_local)

        Parameter
        ---------
        array_local : array-like, optional
            If provided, gather the scattered input array, instead of the whole
            table.

        Returns
        -------
        table_global : PackedTable
            The global packed table, whose all special attribute have been
            MPI-gathered.
        array_global : array
            The MPI-gathered input array.

        """

        def func(x):
            x = np.asarray(x)
            out = np.empty((ntot,) + x.shape[1:], x.dtype)
            nbytes = product(x.shape[1:]) * int(x.itemsize)
            self.comm.Allgatherv(
                x.view(np.byte),
                [
                    out.view(np.byte),
                    ([_ * nbytes for _ in counts], [_ * nbytes for _ in offsets]),
                ],
            )
            return out

        ntot = np.array(len(self))
        self.comm.Allreduce(MPI.IN_PLACE, ntot, op=MPI.SUM)
        counts = []
        offsets = [0]
        for s in split(ntot, self.comm.size):
            n = s.stop - s.start
            counts.append(n)
            offsets.append(offsets[-1] + n)
        offsets.pop()

        if len(args) == 1:
            return func(args[0])
        elif len(args) > 1:
            raise TypeError(f'gather takes at most 1 argument ({len(args)} given)')

        out = copy.copy(self)
        out._index = self._normalize_int_selection(
            func(self.index), product(self.shape[: self.ndim])
        )
        for k in out._special_attributes:
            if k in out._reserved_attributes:
                continue
            try:
                v = self.__dict__[k]
            except KeyError:
                continue
            if isinstance(v, np.ndarray) and v.ndim > 0:
                setattr(out, k, func(v))
        return out

    def split(self, n):
        """
        Split the table in partitioning groups.

        Example
        -------
        >>> table = PackedTable((4, 4), selection=[0, 1, 4, 5])
        >>> print(table.split(2))
        (PackedTable((4, 4), index=slice(0, 2, 1)),
         PackedTable((4, 4), index=slice(4, 6, 1)))

        """
        return tuple(self[_] for _ in split(len(self), n))

    def unpack(self, x, out=None, missing_value=None, copy=False):
        """
        Convert a 1-dimensional array into a multi-dimensional array which
        includes the non-selected components, mimicking the multi-dimensional
        layout.

        Parameters
        ----------
        x : ndarray
            Array to be unpacked, whose first dimension is the number of
            selected components.
        out : ndarray, optional
            Placeholder for the unpacked array.
        missing_value : any, optional
            The value to be used for non-selected components.
        copy : boolean, optional
            Setting this keyword to True ensures that the output is not a view
            of the input x.

        Returns
        -------
        output : ndarray
            Unpacked array, whose first dimensions are equal to those of the
            table attributes.

        See Also
        --------
        pack : inverse method.

        Notes
        -----
        Unless the 'copy' keyword is set to True, this method does not make
        a copy if it simply does a reshape (when all components are selected
        and no ordering is specified).

        """
        if x is None:
            return None
        x = np.atleast_1d(x)
        if x.shape[0] not in (1, len(self)):
            raise ValueError(
                f"Invalid input packed shape '{x.shape}'. The expected first dimension "
                f"is '{len(self)}'."
            )
        index = self._indexable

        unpacked_shape = self._shape_actual + x.shape[1:]
        flat_shape = (self._size_actual,) + x.shape[1:]
        has_out = out is not None
        if has_out:
            if not isinstance(out, np.ndarray):
                raise TypeError('The output array is not an ndarray.')
            if out.shape != unpacked_shape:
                raise ValueError(
                    f"The output array shape '{out.shape}' is invalid. The expected "
                    f"shape is '{unpacked_shape}'."
                )
        else:
            if (
                index is Ellipsis
                and not copy
                and x.shape[0] == len(self)
                and self.comm.size == 1
            ):
                return x.reshape(unpacked_shape)
            out = np.empty(unpacked_shape, dtype=x.dtype).view(type(x))
            if out.__array_finalize__ is not None:
                out.__array_finalize__(x)
        if self._size_actual > len(self):
            if missing_value is None:
                missing_value = self._get_default_missing_value(x.dtype)
            out[...] = missing_value
        elif index is Ellipsis and x.shape[0] == len(self) and self.comm.size == 1:
            out[...] = x.reshape(unpacked_shape)
            return out
        out_ = out.reshape(flat_shape)
        if self.comm.size == 1:
            out_[index] = x
        else:
            ix = self.comm.allgather((index, x))
            for i_, x_ in ix:
                out_[i_] = x_
        if has_out and not isalias(out, out_):
            out[...] = out_.reshape(out.shape)
        return out

    def _get_default_missing_value(self, dtype):
        value = np.empty((), dtype)
        if dtype.kind == 'V':
            for f, (dt, junk) in dtype.fields.items():
                value[f] = self._get_default_missing_value(dt)
            return value
        return {
            'b': False,
            'i': -1,
            'u': 0,
            'f': np.nan,
            'c': np.nan,
            'S': '',
            'U': '',
            'O': None,
        }[dtype.kind]

    def _normalize_selection(self, selection, shape):
        # return the selection as an array of int or bool, a slice, an Ellipsis
        # or a tuple.
        if isinstance(selection, list) and len(selection) == 0:
            return np.array([], self._dtype_index)
        if (
            selection is None
            or selection is Ellipsis
            or isinstance(selection, tuple)
            and len(selection) == 0
        ):
            return Ellipsis
        if isscalarlike(selection):
            return np.asarray([selection], self._dtype_index)
        if isinstance(selection, slice):
            return self._normalize_slice(selection, product(shape))
        if isinstance(selection, (list, tuple)):
            selection_ = np.asarray(selection)
            if selection_.dtype == object:
                if len(selection) > len(shape):
                    raise ValueError('Invalid selection dimensions.')
                selection = tuple(
                    self._normalize_slice(_, s) if isinstance(_, slice) else _
                    for _, s in zip(selection, shape)
                )
                try:
                    return selection[: ilast_is_not(selection, Ellipsis) + 1]
                except ValueError:
                    return Ellipsis
            else:
                selection = selection_
        elif not isinstance(selection, np.ndarray):
            raise TypeError('Invalid selection.')

        if selection.dtype not in (bool, int):
            raise TypeError('Invalid selection.')
        if selection.dtype == int:
            if selection.ndim != 1:
                raise ValueError('Index selection is not 1-dimensional.')
        elif selection.shape != shape:
            raise ValueError('Invalid boolean selection dimensions.')

        return selection

    def _normalize_int_selection(self, s, n):
        # all elements in the integer selection should be positive
        size = s.size
        if size <= 1:
            return s
        start = s[0]
        step = s[1] - s[0]
        if np.all(s == start + np.arange(size) * step):
            if size == n and step == 1:
                return Ellipsis
            else:
                stop = start + size * step
                return slice(start, stop, step)
        return s

    @staticmethod
    def _normalize_slice(s, n):
        step = 1 if s.step is None else s.step
        if step == 0:
            raise ValueError('Invalide stride.')

        start = s.start
        if start is None:
            start = 0 if step > 0 else n - 1
        elif start < 0:
            start = n + start
            if start < 0:
                if step < 0:
                    return slice(0, 0, step)
                start = 0
        elif start >= n:
            if step > 0:
                return slice(n - 1, n - 1, step)
            start = n - 1

        stop = s.stop
        if stop is None:
            stop = n if step > 0 else -1
        elif stop < 0:
            stop = n + stop
        elif stop > n:
            stop = n

        if step > 0:
            extent = max(stop - start, 0)
        else:
            extent = max(start - stop, 0)
        size = int(np.ceil(extent / abs(step)))
        if size == n and step == 1:
            return Ellipsis

        return slice(start, start + size * step, step)

    def _selection2bool(self, selection):
        out = np.zeros(self._shape_actual, bool)
        out[selection] = True
        return out


class UnpackedTable:
    def __init__(self, packed):
        object.__setattr__(self, '_packed', packed)
        for k in packed._special_attributes:
            if k == 'removed':
                continue
            object.__setattr__(self, k, None)  # for IPython autocompletion

    @property
    def removed(self):
        out = self._packed.unpack(False, missing_value=True)
        out.flags.writeable = False
        return out

    def __len__(self):
        return product(self._packed.shape)

    def __getattribute__(self, key):
        if key.startswith('_') or key == 'removed':
            return object.__getattribute__(self, key)
        if key not in self._packed._special_attributes:
            raise AttributeError(f'The unpacked table has no attribute {key!r}.')
        v = getattr(self._packed, key)
        if v is None:
            return v
        if isinstance(v, Callable):
            return lambda *args: self._packed.unpack(v(*args))
        out = self._packed.unpack(v).view()
        out.flags.writeable = False
        return out

    #    def __getitem__(self, selection):
    #        shape = self._packed.shape
    #        selection = self._packed._normalize_selection(selection, shape)
    #        if selection is Ellipsis:
    #            return self
    #        selection = self._packed._selection2bool(selection)
    #        index = self._packed.index
    #        return self._packed[selection[index]]

    def __setattr__(self, key, value):
        p = self._packed
        if key not in p._special_attributes:
            raise AttributeError(f'{key!r} is not an unpacked attribute.')
        if key in p._reserved_attributes:
            raise AttributeError(f'The attribute {key!r} is not writeable.')
        if isinstance(value, Callable):
            raise TypeError('A function cannot be set as an unpacked attribute.')
        object.__setattr__(p, key, p.pack(value, copy=True))



from pathlib import Path

import astropy.io.fits as pyfits
import numpy as np
import pytest
from numpy.testing import assert_allclose

from pyoperators import DiagonalOperator, PackOperator, pcg
from pyoperators.utils.testing import assert_eq, assert_same
from pysimulators import Map
from pysimulators.interfaces.madmap1 import MadMap1Observation, _read_filters

DATAPATH = Path(__file__).parent / 'data' / 'madmap1'


def mapper_naive(tod, H):
    m, c = H.T(tod), H.T(np.ones_like(tod))
    m /= c
    m.coverage = c
    return m


@pytest.mark.xfail(reason='reason: incorrect coverage in mapper_naive')
def test_ls():
    profile = None  #'test_madcap.png'
    map_ref = pyfits.open(DATAPATH / 'naivemapSpirePSW.fits')['image'].data
    name = 'SPIRE/PSW'
    obs = MadMap1Observation(
        name,
        135,
        str(DATAPATH / 'todSpirePSW_be'),
        str(DATAPATH / 'invnttSpirePSW_be'),
        str(DATAPATH / 'madmapSpirePSW.fits[coverage]'),
        bigendian=True,
        missing_value=np.nan,
    )

    tod = obs.get_tod(unit='Jy/beam')
    projection = obs.get_operator()
    packing = PackOperator(np.isfinite(map_ref))

    map_naive_1d = mapper_naive(tod, projection)
    assert_same(map_naive_1d, map_ref[np.isfinite(map_ref)])

    model = projection * packing
    map_naive_2d = mapper_naive(tod, model)
    assert_same(map_naive_2d, map_ref)

    M = DiagonalOperator(packing(1 / map_naive_2d.coverage))
    assert np.all(np.isfinite(M.data))

    invntt = obs.get_invntt_operator(fftw_flag='FFTW_PATIENT')

    class Callback:
        def __init__(self):
            self.niterations = 0

        def __call__(self, x):
            self.niterations += 1

    callback = Callback()
    # callback = None

    H = projection
    m = pcg(H.T * invntt * H, (H.T * invntt)(tod), M=M, tol=1e-7)
    map_ls_packed = Map(m['x'])
    map_ls_packed.header['TIME'] = m['time']

    # print('Elapsed time:', map_ls_packed.header['TIME'])
    # from tamasis import mapper_ls
    # map_ls_packed = mapper_ls(tod, projection, invntt=invntt, tol=1e-7, M=M,
    #                          callback=callback, criterion=False,
    #                          profile=profile)
    if profile:
        return
    print('Elapsed time:', map_ls_packed.header['TIME'])
    assert m['nit'] < 50
    ref = packing(Map(str(DATAPATH / 'madmapSpirePSW.fits')))
    assert_allclose(map_ls_packed, ref, atol=1e-5)


def test_endian():
    bendian = _read_filters(DATAPATH / 'invntt_be', bigendian=True)
    lendian = _read_filters(DATAPATH / 'invntt_le')
    assert len(bendian) == 6
    assert bendian[0]['data'].size == 101
    assert_same(bendian[0]['data'][0], 5597147.4155586753)
    assert_eq(lendian, bendian)



try:
    import healpy as hp
    import healpy._healpy_pixel_lib as pixlib
except ImportError:
    hp = None
    pixlib = None
import numpy as np

from pyoperators import CompositionOperator, IdentityOperator, Operator
from pyoperators.flags import inplace, real, square, symmetric
from pyoperators.utils import pi, strenum

from ...sparse import FSRMatrix, SparseOperator

__all__ = [
    'Healpix2CartesianOperator',
    'Cartesian2HealpixOperator',
    'Healpix2SphericalOperator',
    'Spherical2HealpixOperator',
    'HealpixConvolutionGaussianOperator',
    'HealpixLaplacianOperator',
]


@real
class _HealPixCartesian(Operator):
    def __init__(self, nside, nest=False, dtype=float, **keywords):
        if hp is None:
            raise ImportError('The package healpy is not installed.')
        self.nside = int(nside)
        self.nest = bool(nest)
        Operator.__init__(self, dtype=dtype, **keywords)

    @staticmethod
    def _reshapehealpix(shape):
        return shape + (3,)

    @staticmethod
    def _reshapecartesian(shape):
        return shape[:-1]

    @staticmethod
    def _validatecartesian(shape):
        if len(shape) == 0 or shape[-1] != 3:
            raise ValueError('Invalid cartesian shape.')

    @staticmethod
    def _rule_identity(o1, o2):
        if o1.nside == o2.nside and o1.nest == o2.nest:
            return IdentityOperator()


class Healpix2CartesianOperator(_HealPixCartesian):
    """
    Convert Healpix pixels into cartesian coordinates.

    """

    def __init__(self, nside, nest=False, **keywords):
        """
        nside : int
            Value of the map resolution parameter.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            ring scheme.

        """
        super().__init__(
            nside,
            nest=nest,
            reshapein=self._reshapehealpix,
            reshapeout=self._reshapecartesian,
            validateout=self._validatecartesian,
            **keywords,
        )
        self.set_rule('I', lambda s: Cartesian2HealpixOperator(s.nside, nest=s.nest))
        self.set_rule(
            ('.', Cartesian2HealpixOperator), self._rule_identity, CompositionOperator
        )

    def direct(self, input, output):
        input = input.astype(int)
        func = pixlib._pix2vec_nest if self.nest else pixlib._pix2vec_ring
        func(self.nside, input, output[..., 0], output[..., 1], output[..., 2])


class Cartesian2HealpixOperator(_HealPixCartesian):
    """
    Convert cartesian coordinates into Healpix pixels.

    """

    def __init__(self, nside, nest=False, **keywords):
        """
        nside : int
            Value of the map resolution parameter.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            ring scheme.

        """
        super().__init__(
            nside,
            nest=nest,
            reshapein=self._reshapecartesian,
            reshapeout=self._reshapehealpix,
            validatein=self._validatecartesian,
            **keywords,
        )
        self.set_rule('I', lambda s: Healpix2CartesianOperator(s.nside, nest=s.nest))
        self.set_rule(
            ('.', Healpix2CartesianOperator), self._rule_identity, CompositionOperator
        )

    def direct(self, input, output):
        func = pixlib._vec2pix_nest if self.nest else pixlib._vec2pix_ring
        func(self.nside, input[..., 0], input[..., 1], input[..., 2], output)


@real
class _HealPixSpherical(Operator):
    CONVENTIONS = (
        'zenith,azimuth',
        'azimuth,zenith',
        'elevation,azimuth',
        'azimuth,elevation',
    )

    def __init__(self, nside, convention, nest=False, dtype=float, **keywords):
        if hp is None:
            raise ImportError('The package healpy is not installed.')
        if not isinstance(convention, str):
            raise TypeError(f"The input convention '{convention}' is not a string.")
        convention_ = convention.replace(' ', '').lower()
        if convention_ not in self.CONVENTIONS:
            raise ValueError(
                f'Invalid spherical convention {convention!r}. Expected values are '
                f'{strenum(self.CONVENTIONS)}.'
            )
        self.nside = int(nside)
        self.convention = convention_
        self.nest = bool(nest)
        Operator.__init__(self, dtype=dtype, **keywords)

    @staticmethod
    def _reshapehealpix(shape):
        return shape + (2,)

    @staticmethod
    def _reshapespherical(shape):
        return shape[:-1]

    @staticmethod
    def _validatespherical(shape):
        if len(shape) == 0 or shape[-1] != 2:
            raise ValueError('Invalid spherical shape.')

    @staticmethod
    def _rule_identity(o1, o2):
        if (
            o1.nside == o2.nside
            and o1.convention == o2.convention
            and o1.nest == o2.nest
        ):
            return IdentityOperator()


class Healpix2SphericalOperator(_HealPixSpherical):
    """
    Convert Healpix pixels into spherical coordinates in radians.


    The last dimension of the operator's output is 2 and it encodes
    the two spherical angles. Four conventions define what these angles are:
       - 'zenith,azimuth': (theta, phi) angles commonly used
       in physics or the (colatitude, longitude) angles used
       in the celestial and geographical coordinate systems
       - 'azimuth,zenith': (longitude, colatitude) convention
       - 'elevation,azimuth: (latitude, longitude) convention
       - 'azimuth,elevation': (longitude, latitude) convention

    """

    def __init__(self, nside, convention, nest=False, **keywords):
        """
        nside : int
            Value of the map resolution parameter.
        convention : string
            One of the following spherical coordinate conventions:
            'zenith,azimuth', 'azimuth,zenith', 'elevation,azimuth' and
            'azimuth,elevation'.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            ring scheme.

        """
        super().__init__(
            nside,
            convention,
            nest=nest,
            reshapein=self._reshapehealpix,
            reshapeout=self._reshapespherical,
            validateout=self._validatespherical,
            **keywords,
        )
        self.set_rule(
            'I', lambda s: Spherical2HealpixOperator(s.nside, s.convention, nest=s.nest)
        )
        self.set_rule(
            ('.', Spherical2HealpixOperator), self._rule_identity, CompositionOperator
        )

    def direct(self, input, output):
        input = input.astype(int)
        func = pixlib._pix2ang_nest if self.nest else pixlib._pix2ang_ring
        if self.convention.startswith('azimuth'):
            o1, o2 = output[..., 1], output[..., 0]
        else:
            o1, o2 = output[..., 0], output[..., 1]
        func(self.nside, input, o1, o2)
        if 'elevation' in self.convention:
            np.subtract(0.5 * pi(self.dtype), o1, o1)


class Spherical2HealpixOperator(_HealPixSpherical):
    """
    Convert spherical coordinates in radians into Healpix pixels.

    The last dimension of the operator's output is 2 and it encodes
    the two spherical angles. Four conventions define what these angles are:
       - 'zenith,azimuth': (theta, phi) angles commonly used
       in physics or the (colatitude, longitude) angles used
       in the celestial and geographical coordinate systems
       - 'azimuth,zenith': (longitude, colatitude) convention
       - 'elevation,azimuth: (latitude, longitude) convention
       - 'azimuth,elevation': (longitude, latitude) convention

    """

    def __init__(self, nside, convention, nest=False, **keywords):
        """
        nside : int
            Value of the map resolution parameter.
        convention : string
            One of the following spherical coordinate conventions:
            'zenith,azimuth', 'azimuth,zenith', 'elevation,azimuth' and
            'azimuth,elevation'.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            ring scheme.

        """
        super().__init__(
            nside,
            convention,
            nest=nest,
            reshapein=self._reshapespherical,
            reshapeout=self._reshapehealpix,
            validatein=self._validatespherical,
            **keywords,
        )
        self.set_rule(
            'I', lambda s: Healpix2SphericalOperator(s.nside, s.convention, nest=s.nest)
        )
        self.set_rule(
            ('.', Healpix2SphericalOperator), self._rule_identity, CompositionOperator
        )

    def direct(self, input, output):
        func = pixlib._ang2pix_nest if self.nest else pixlib._ang2pix_ring
        if self.convention.startswith('azimuth'):
            i1, i2 = input[..., 1], input[..., 0]
        else:
            i1, i2 = input[..., 0], input[..., 1]
        if 'elevation' in self.convention:
            i1 = 0.5 * pi(self.dtype) - i1
        func(self.nside, i1, i2, output)


@inplace
@real
@square
@symmetric
class HealpixConvolutionGaussianOperator(Operator):
    """
    Convolve a Healpix map by a gaussian kernel.

    """

    def __init__(
        self,
        fwhm=None,
        sigma=None,
        iter=3,
        lmax=None,
        mmax=None,
        use_weights=False,
        datapath=None,
        dtype=float,
        pol=True,
        **keywords,
    ):
        """
        Keywords are passed to the Healpy function smoothing.

        """
        if hp is None:
            raise ImportError('The package healpy is not installed.')
        if fwhm is None and sigma is None:
            raise ValueError('The convolution width is not specified.')
        if fwhm is not None and sigma is not None:
            raise ValueError('Ambiguous convolution width specification.')
        if fwhm is not None and fwhm < 0 or sigma is not None and sigma < 0:
            raise ValueError('The convolution width is not positive.')
        if fwhm in (0, None) and sigma in (0, None):
            self.__class__ = IdentityOperator
            self.__init__(**keywords)
            return
        if fwhm is None:
            fwhm = sigma * np.sqrt(8 * np.log(2))
        if sigma is None:
            sigma = fwhm / np.sqrt(8 * np.log(2))
        self.fwhm = fwhm
        self.sigma = sigma
        self.iter = iter
        self.lmax = lmax
        self.mmax = mmax
        self.use_weights = use_weights
        self.datapath = datapath
        self.pol = pol
        Operator.__init__(self, dtype=dtype, **keywords)

    def direct(self, input, output):
        if input.ndim > 1:
            input = [_ for _ in input.T]
        output.T[...] = hp.smoothing(
            input,
            fwhm=self.fwhm,
            iter=self.iter,
            lmax=self.lmax,
            mmax=self.mmax,
            use_weights=self.use_weights,
            datapath=self.datapath,
            pol=self.pol,
        )

    def validatein(self, shape):
        if len(shape) == 0 or len(shape) > 2:
            raise ValueError('Invalid number of dimensions.')
        nside = int(np.round(np.sqrt(shape[0] / 12)))
        if 12 * nside**2 != shape[0]:
            raise ValueError(
                f'The nside value cannot be inferred from the input number of pixels '
                f"'{shape[0]}'."
            )


@real
@square
@symmetric
class HealpixLaplacianOperator(SparseOperator):
    """
    9-Point stencil laplacian for Healpix maps.

    """

    def __init__(self, nside):
        """
        Parameters
        ----------
        nside : integer
            The Healpix map nside.

        """
        if hp is None:
            raise ImportError('The package healpy is not installed.')
        npix = 12 * nside**2
        ipix = np.arange(npix, dtype=np.int32)
        neighbours = hp.get_all_neighbours(nside, ipix)
        s = FSRMatrix((npix, npix), ncolmax=9, dtype_index=np.int32)
        s.data.index[:, 0] = ipix
        s.data.index[:, 1:] = neighbours.T
        h2 = 4 * np.pi / npix
        s.data.value[:, 0] = -20 / (6 * h2)
        s.data.value[:, 1:] = np.array([1, 4, 1, 4, 1, 4, 1, 4]) / (6 * h2)
        self.nside = nside
        SparseOperator.__init__(self, s)



import numpy as np
import pytest
from numpy.testing import assert_equal

from pyoperators.utils.testing import assert_same
from pysimulators import Layout, LayoutGrid, LayoutGridSquares, Quantity
from pysimulators.geometry import create_grid, create_grid_squares


def test_layout_grid_errors():
    with pytest.raises(TypeError):
        LayoutGrid(3, 1.2)
    with pytest.raises(ValueError):
        LayoutGrid((1,), 1.2)
    with pytest.raises(ValueError):
        LayoutGrid((1, 2, 3), 1.2)
    with pytest.raises(ValueError):
        LayoutGrid((4, 2), 0.1, origin=[1])


LAYOUT_SHAPE = (4, 4)
LAYOUT_CENTERS = create_grid(LAYOUT_SHAPE, 0.1)


class LayoutProperty(Layout):
    @property
    def center(self):
        return LAYOUT_CENTERS.reshape(-1, 2)


class LayoutMethod(Layout):
    def center(self):
        return LAYOUT_CENTERS.reshape(-1, 2)


@pytest.mark.parametrize(
    'cls, center',
    [
        (Layout, LAYOUT_CENTERS),
        (Layout, lambda: LAYOUT_CENTERS.reshape(-1, 2)),
        (LayoutProperty, None),
        (LayoutMethod, None),
    ],
)
def test_layout(cls, center):
    layout = cls(LAYOUT_SHAPE, center=center)
    assert_same(layout.center, LAYOUT_CENTERS.reshape(-1, 2))
    assert_same(layout.all.center, LAYOUT_CENTERS)


LAYOUT_VERTICES = create_grid_squares(LAYOUT_SHAPE, 0.1, filling_factor=0.8)


class LayoutVertexProperty(Layout):
    @property
    def vertex(self):
        return LAYOUT_VERTICES.reshape(-1, 4, 2)


class LayoutVertexMethod(Layout):
    def vertex(self):
        return LAYOUT_VERTICES.reshape(-1, 4, 2)


@pytest.mark.parametrize(
    'cls, vertex',
    [
        (Layout, LAYOUT_VERTICES),
        (Layout, lambda: LAYOUT_VERTICES.reshape(-1, 4, 2)),
        (LayoutVertexProperty, None),
        (LayoutVertexMethod, None),
    ],
)
def test_layout_vertex(cls, vertex):
    layout = cls(LAYOUT_SHAPE, vertex=vertex)
    center = np.mean(LAYOUT_VERTICES, axis=-2)

    assert layout.nvertices == 4
    assert_same(layout.center, center.reshape((-1, 2)))
    assert_same(layout.all.center, center)
    assert_same(layout.vertex, LAYOUT_VERTICES.reshape((-1, 4, 2)))
    assert_same(layout.all.vertex, LAYOUT_VERTICES)


def test_layout_grid():
    shape = (3, 2)
    spacing = Quantity(0.1, 'mm')
    xreflection = True
    yreflection = True
    origin = (1, 1)
    angle = 10
    center = create_grid(
        shape,
        spacing,
        center=origin,
        xreflection=xreflection,
        yreflection=yreflection,
        angle=angle,
    )
    layout = LayoutGrid(
        shape,
        spacing,
        origin=origin,
        xreflection=xreflection,
        yreflection=yreflection,
        angle=angle,
    )
    assert not hasattr(layout, 'nvertices')
    assert_same(layout.all.center, center)
    assert_same(layout.all.removed, np.zeros(shape, bool))


def test_layout_grid_circles():
    shape = (3, 2)
    spacing = 0.1
    xreflection = True
    yreflection = True
    origin = (1, 1)
    angle = 10
    center = create_grid(
        shape,
        spacing,
        center=origin,
        xreflection=xreflection,
        yreflection=yreflection,
        angle=angle,
    )
    layout = LayoutGrid(
        shape,
        spacing,
        origin=origin,
        xreflection=xreflection,
        yreflection=yreflection,
        angle=angle,
        radius=spacing / 2,
    )
    assert not hasattr(layout, 'nvertices')
    assert_same(layout.radius, spacing / 2)
    assert_same(layout.all.center, center)
    assert_same(layout.all.removed, np.zeros(shape, bool))


LAYOUT_GRID_CIRCLES_UNIT_SHAPE = (3, 2)


@pytest.mark.parametrize('spacing', [1.0, Quantity(1.0), Quantity(1.0, 'mm')])
@pytest.mark.parametrize(
    'radius',
    [
        0.4,
        Quantity(0.4),
        Quantity(0.4, 'mm'),
        Quantity(0.4e-3, 'm'),
        np.zeros(LAYOUT_GRID_CIRCLES_UNIT_SHAPE) + 0.4,
        Quantity(np.zeros(LAYOUT_GRID_CIRCLES_UNIT_SHAPE) + 0.4),
        Quantity(np.zeros(LAYOUT_GRID_CIRCLES_UNIT_SHAPE) + 0.4, 'mm'),
        Quantity((np.zeros(LAYOUT_GRID_CIRCLES_UNIT_SHAPE) + 0.4) * 1e-3, 'm'),
    ],
)
def test_layout_grid_circles_unit(spacing, radius):
    selection = [[True, False], [True, True], [True, True]]

    layout = LayoutGrid(
        LAYOUT_GRID_CIRCLES_UNIT_SHAPE, spacing, radius=radius, selection=selection
    )
    if (
        (not isinstance(spacing, Quantity) or spacing.unit == '')
        and isinstance(radius, Quantity)
        and radius.unit == 'm'
    ):
        expectedval = 0.4e-3
    else:
        expectedval = 0.4
    expected = np.empty(LAYOUT_GRID_CIRCLES_UNIT_SHAPE)
    expected[...] = expectedval
    expected[0, 1] = np.nan
    assert_same(layout.radius, expectedval, broadcasting=True)
    assert_same(layout.all.radius, expected)
    vals = (layout.radius, layout.all.radius, layout.center, layout.all.center)
    unit = getattr(spacing, 'unit', '') or getattr(radius, 'unit', '')
    if unit:
        for val in vals:
            assert isinstance(val, Quantity)
            assert val.unit == unit
    else:
        radius = np.asanyarray(radius)
        expecteds = (type(radius), type(radius), np.ndarray, np.ndarray)
        for val, e in zip(vals, expecteds):
            assert type(val) is e


def test_layout_grid_squares():
    shape = (3, 2)
    spacing = 0.1
    filling_factor = 0.9
    xreflection = True
    yreflection = True
    origin = (1, 1)
    angle = 10
    vertex = create_grid_squares(
        shape,
        spacing,
        filling_factor=filling_factor,
        center=origin,
        xreflection=xreflection,
        yreflection=yreflection,
        angle=angle,
    )
    layout = LayoutGridSquares(
        shape,
        spacing,
        filling_factor=filling_factor,
        origin=origin,
        xreflection=xreflection,
        yreflection=yreflection,
        angle=angle,
    )
    assert layout.nvertices == 4
    assert_same(layout.all.center, np.mean(vertex, axis=-2))
    assert_same(layout.all.vertex, vertex)
    assert_same(layout.all.removed, np.zeros(shape, bool))


@pytest.mark.parametrize('spacing', [1.0, Quantity(1.0), Quantity(1.0, 'mm')])
@pytest.mark.parametrize(
    'lcenter',
    [
        (1.0, 1.0),
        Quantity((1.0, 1.0)),
        Quantity((1.0, 1.0), 'mm'),
        Quantity((1e-3, 1e-3), 'm'),
    ],
)
def test_layout_grid_squares_unit(spacing, lcenter):
    shape = (3, 2)
    selection = [[True, True], [True, True], [True, True]]

    layout = LayoutGridSquares(shape, spacing, selection=selection, origin=lcenter)
    if (
        (not isinstance(spacing, Quantity) or spacing.unit == '')
        and isinstance(lcenter, Quantity)
        and lcenter.unit == 'm'
    ):
        expected = np.array((1e-3, 1e-3))
    else:
        expected = np.array((1, 1))
    actual = np.mean(layout.center, axis=0).view(np.ndarray)
    assert_same(actual, expected, rtol=1000)
    vals = (layout.center, layout.all.center, layout.vertex, layout.all.vertex)
    unit = getattr(spacing, 'unit', '') or getattr(lcenter, 'unit', '')
    if unit:
        for val in vals:
            assert isinstance(val, Quantity)
            assert val.unit == unit
    else:
        for val in vals:
            assert type(val) is np.ndarray


def test_layout_grid_colrow():
    ordering = [[-1, 3, 0], [1, -1, -1], [-1, 2, -1]]
    expected_row = [[-1, 0, 0], [1, -1, -1], [-1, 2, -1]]
    expected_col = [[-1, 1, 2], [0, -1, -1], [-1, 1, -1]]
    grid = LayoutGrid((3, 3), 1.2, ordering=ordering)
    assert_equal(grid.all.column, expected_col)
    assert_equal(grid.all.row, expected_row)



from itertools import chain, product

import numpy as np
import pytest

from pyoperators import MPI, MPIDistributionIdentityOperator
from pyoperators.utils.testing import assert_same
from pysimulators import Acquisition, Instrument, PackedTable, Sampling, Scene
from pysimulators.operators import ProjectionOperator
from pysimulators.sparse import FSRMatrix

pytestmark = pytest.mark.mpi

RANK = MPI.COMM_WORLD.rank
SIZE = MPI.COMM_WORLD.size
NPROCS_INSTRUMENT = sorted(
    {int(n) for n in [1, SIZE / 3, SIZE / 2, SIZE] if int(n) == n}
)
NSCENE = 10
NSAMPLING_GLOBAL = 100
NDETECTOR_GLOBAL = 16
SCENE = Scene(NSCENE)
SAMPLING = Sampling(NSAMPLING_GLOBAL, period=1.0)
INSTRUMENT = Instrument('', PackedTable(NDETECTOR_GLOBAL))


class MyAcquisition(Acquisition):
    def get_projection_operator(self):
        dtype = [('index', int), ('value', float)]
        data = np.recarray((len(self.instrument), len(self.sampling), 1), dtype=dtype)
        for ilocal, iglobal in enumerate(self.instrument.detector.index):
            data[ilocal].value = iglobal
            data[ilocal, :, 0].index = [
                (iglobal + int(t)) % NSCENE for t in self.sampling.time
            ]

        matrix = FSRMatrix(
            (len(self.instrument) * len(self.sampling), NSCENE),
            data=data.reshape((-1, 1)),
        )

        return ProjectionOperator(
            matrix, dtype=float, shapeout=(len(self.instrument), len(self.sampling))
        )


def get_acquisition(comm, nprocs_instrument):
    return MyAcquisition(
        INSTRUMENT, SAMPLING, SCENE, comm=comm, nprocs_instrument=nprocs_instrument
    )


@pytest.mark.parametrize('nprocs_instrument', NPROCS_INSTRUMENT)
def test_communicators(nprocs_instrument):
    sky = SCENE.ones()
    nprocs_sampling = SIZE // nprocs_instrument
    serial_acq = get_acquisition(MPI.COMM_SELF, 1)
    assert serial_acq.comm.size == 1
    assert serial_acq.instrument.comm.size == 1
    assert serial_acq.sampling.comm.size == 1
    assert len(serial_acq.instrument) == NDETECTOR_GLOBAL
    assert len(serial_acq.sampling) == NSAMPLING_GLOBAL

    parallel_acq = get_acquisition(MPI.COMM_WORLD, nprocs_instrument)
    assert parallel_acq.comm.size == SIZE
    assert parallel_acq.instrument.comm.size == nprocs_instrument
    assert parallel_acq.sampling.comm.size == nprocs_sampling
    assert (
        parallel_acq.instrument.comm.allreduce(len(parallel_acq.instrument))
        == NDETECTOR_GLOBAL
    )
    assert (
        parallel_acq.sampling.comm.allreduce(len(parallel_acq.sampling))
        == NSAMPLING_GLOBAL
    )

    serial_H = serial_acq.get_projection_operator()
    ref_tod = serial_H(sky)

    parallel_H = (
        parallel_acq.get_projection_operator()
        * MPIDistributionIdentityOperator(parallel_acq.comm)
    )
    local_tod = parallel_H(sky)
    actual_tod = np.vstack(
        parallel_acq.instrument.comm.allgather(
            np.hstack(parallel_acq.sampling.comm.allgather(local_tod))
        )
    )
    assert_same(actual_tod, ref_tod, atol=20)

    ref_backproj = serial_H.T(ref_tod)
    actual_backproj = parallel_H.T(local_tod)
    assert_same(actual_backproj, ref_backproj, atol=20)


@pytest.mark.parametrize('nprocs_instrument', NPROCS_INSTRUMENT)
@pytest.mark.parametrize(
    'selection',
    [
        Ellipsis,
        slice(None),
    ]
    + list(chain(*(product([slice(None), Ellipsis], repeat=n) for n in [1, 2, 3]))),
)
def test_communicators_getitem_all(nprocs_instrument, selection):
    acq = get_acquisition(MPI.COMM_WORLD, nprocs_instrument)
    assert acq.instrument.comm.size == nprocs_instrument
    assert acq.sampling.comm.size == MPI.COMM_WORLD.size / nprocs_instrument
    assert acq.comm.size == MPI.COMM_WORLD.size
    restricted_acq = acq[selection]
    assert restricted_acq.instrument.comm.size == nprocs_instrument
    assert restricted_acq.sampling.comm.size == MPI.COMM_WORLD.size / nprocs_instrument
    assert restricted_acq.comm.size == MPI.COMM_WORLD.size


@pytest.mark.parametrize('nprocs_instrument', NPROCS_INSTRUMENT)
@pytest.mark.parametrize('selection', [0, slice(None, 1), np.array])
def test_communicators_getitem_instrument(nprocs_instrument, selection):
    acq = get_acquisition(MPI.COMM_WORLD, nprocs_instrument)
    if selection is np.array:
        selection = np.zeros(len(acq.instrument), bool)
        selection[0] = True
    restricted_acq = acq[selection]
    assert restricted_acq.instrument.comm.size == 1
    assert restricted_acq.sampling.comm.size == acq.sampling.comm.size
    assert restricted_acq.comm.size == acq.sampling.comm.size


SELECTION_GETITEM_SAMPLING = np.zeros(NSAMPLING_GLOBAL, bool)
SELECTION_GETITEM_SAMPLING[0] = True


@pytest.mark.parametrize('nprocs_instrument', NPROCS_INSTRUMENT)
@pytest.mark.parametrize('selection', [0, slice(None, 1), np.array])
def test_communicators_getitem_sampling(nprocs_instrument, selection):
    acq = get_acquisition(MPI.COMM_WORLD, nprocs_instrument)
    if selection is np.array:
        selection = np.zeros(len(acq.sampling), bool)
        selection[0] = True
    restricted_acq = acq[:, selection]
    assert restricted_acq.instrument.comm.size == acq.instrument.comm.size
    assert restricted_acq.sampling.comm.size == 1
    assert restricted_acq.comm.size == acq.instrument.comm.size



import numpy as np

from pyoperators import (
    IdentityOperator,
    MPIDistributionIdentityOperator,
    To1dOperator,
    asoperator,
)
from pyoperators.memory import empty
from pyoperators.utils import product, strenum

from .. import _flib as flib
from ..operators import ProjectionOperator
from ..sparse import FSRMatrix
from ..wcsutils import WCSToPixelOperator, fitsheader2shape
from .core import PackedTable

__all__ = ['Scene', 'SceneGrid']


class Scene(PackedTable):
    """
    Class to handle plane pixelisation and 1-d indexing.

    """

    def __init__(self, shape, topixel=None, ndim=None, dtype=float, **keywords):
        """
        Parameters
        ----------
        shape : tuple of integers
            The surface shape.
        ndim : int, optional
            The number of splittable (indexable) dimensions. It is the actual
            number of dimensions of the layout. It can be lower than that
            specified by the layout shape, in which case the extra dimensions
            are instructed not to be split.
        topixel : Operator, optional
            World-to-pixel coordinate transform.

        """
        PackedTable.__init__(self, shape, ndim=ndim)
        if topixel is None:
            topixel = IdentityOperator(self.shape[: self.ndim])
        else:
            topixel = asoperator(topixel)
        self.dtype = np.dtype(dtype)
        self.topixel = topixel
        self.toworld = topixel.I
        for k, v in keywords.items():
            setattr(self, k, v)

    def empty(self):
        """
        Return a new array as described by the scene, without initializing
        entries.

        """
        return empty(self.shape, self.dtype)

    def ones(self):
        """
        Return a new array as described by the scene, filled with ones.

        """
        out = self.empty()
        out[...] = 1
        return out

    def zeros(self):
        """
        Return a new array as described by the scene, filled with zeros.

        """
        out = self.empty()
        out[...] = 0
        return out

    def get_distribution_operator(self, comm):
        """
        Distribute a global scene, of which each MPI process has a copy, to the
        MPI processes.

        It is a block column operator whose blocks are identities distributed
        across the MPI processes.

                       |1   O|
        MPI rank 0 --> |  .  |
                       |O   1|
                       +-----+
                       |1   O|
        MPI rank 1 --> |  .  |
                       |O   1|
                       +-----+
                       |1   O|
        MPI rank 2 --> |  .  |
                       |O   1|

        For an MPI process, the direct method is the Identity and the transpose
        method is a reduction.

        """
        if self.comm.size > 1:
            raise ValueError('The scene is already distributed.')
        return MPIDistributionIdentityOperator(comm)


class SceneGrid(Scene):
    """
    Class for 2-dimensional scenes.

    """

    def __init__(
        self,
        shape,
        topixel=None,
        to1d=None,
        origin='upper',
        startswith1=False,
        **keywords,
    ):
        """
        Parameters
        ----------
        shape : tuple of integers
            The surface shape.
        startswith1 : bool
            If True, columns and row starts with 1 instead of 0.
        topixel : Operator, optional
            World-to-pixel coordinate transform.
        to1d : Operator, optional
            Nd-to-1d pixel index transform.

        """
        origins = 'upper', 'lower'
        if not isinstance(origin, str):
            raise TypeError('Invalid origin.')
        origin = origin.lower()
        if origin not in origins:
            raise ValueError(
                f'Invalid origin {origin!r}. Expected values are {strenum(origins)}.'
            )
        Scene.__init__(self, shape, topixel=topixel, **keywords)
        if self.ndim != 2:
            raise ValueError('The scene is not 2-dimensional.')
        self.origin = origin
        self.startswith1 = bool(startswith1)
        if to1d is not None:
            to1d = asoperator(to1d)
            self.toNd = to1d.I
        else:
            self.toNd = None
        self.to1d = to1d

    @classmethod
    def fromfits(cls, header, **keywords):
        """
        Return a Scene grid described by the WCS of a FITS header.

        Parameters
        ----------
        header : astropy.io.fits.Header
            The FITS header that defines the WCS transform.

        """
        shape = fitsheader2shape(header)
        topixel = WCSToPixelOperator(header)
        to1d = To1dOperator(shape[::-1], order='f')
        scene = cls(shape, topixel, to1d, origin='lower', **keywords)
        scene.header = header
        return scene

    def column(self):
        return self.index % self.shape[1] + self.startswith1

    def row(self):
        return self.index // self.shape[1] + self.startswith1

    def get_integration_operator(
        self, polygons, ncolmax=0, dtype=np.float64, dtype_index=None, **keywords
    ):
        """
        Return the Projection Operator that spatially integrates
        the plane pixel values enclosed by a sequence of polygons.

        Parameters
        ----------
        polygons : array-like of shape (..., nvertices, 2)
            The vertices defining the polygons inside which the spatial
            integration is done, in pixel coordinate units.
        ncolmax : integer
            The maximum number of surface pixel intersected by the polygons.
            If set to zero, or less than the required value, a two-pass
            computation will be performed to determine it.
        dtype : dtype, optional
            The float datatype of the matrix elements.
        dtype_index : dtype, optional
            The integer datatype of the matrix. The defaultis np.int32 or
            np.int64 depending on the number of elements of the scene.

        """
        if len(self) != len(self.all):
            raise NotImplementedError('The scene has removed elements.')
        dtype = np.dtype(dtype)
        if dtype_index is None:
            if len(self) <= np.iinfo(np.int32).max:
                dtype_index = np.int32
            else:
                dtype_index = np.int64
        dtype_index = np.dtype(dtype_index)
        if str(dtype_index) not in ('int32', 'int64') or str(dtype) not in (
            'float32',
            'float64',
        ):
            raise TypeError(
                f'The projection matrix cannot be created with an index type '
                f'{dtype_index} and a value type {dtype}'
            )

        polygons = np.asarray(polygons, dtype=dtype)
        shape = polygons.shape[:-2]
        matrix = FSRMatrix(
            (product(shape), product(self.shape)),
            dtype=dtype,
            dtype_index=dtype_index,
            ncolmax=ncolmax,
        )

        if matrix.data.size == 0:
            # f2py doesn't accept zero-sized opaque arguments
            data = np.empty(1, np.int8)
        else:
            data = matrix.data.ravel().view(np.int8)

        nvertices = polygons.shape[-2]
        isize = dtype_index.itemsize
        rsize = dtype.itemsize
        func = f'matrix_polygon_integration_i{isize}_r{rsize}'
        min_ncolmax, outside = getattr(flib.projection, func)(
            polygons.reshape(-1, nvertices, 2).T,
            self.shape[1],
            self.shape[0],
            data,
            ncolmax,
        )

        if min_ncolmax > ncolmax:
            return self.get_integration_operator(
                polygons,
                ncolmax=min_ncolmax,
                dtype=dtype,
                dtype_index=dtype_index,
                **keywords,
            )

        out = ProjectionOperator(matrix, shapein=self.shape, shapeout=shape, **keywords)
        out.method = 'polygon integration'
        out.outside = bool(outside)
        out.min_ncolmax = min_ncolmax
        return out
